import abc
from typing import Optional, Sequence

import torch
import torch.distributions as dist
from tensordict import TensorDict


class SEM(dist.Distribution, abc.ABC):
    arg_constraints: dict = {}

    def __init__(
        self,
        graph: torch.Tensor,
        node_names: Sequence[str],
        event_shape: torch.Size = torch.Size(),
        batch_shape: torch.Size = torch.Size(),
    ) -> None:
        """
        Args:
            graph: Adjacency Matrix for the underlying SEM
            node_names: The names of each node in the graph
            event_shape: Shape of a sample from the distribution
            batch_shape: Batch shape of the distribution
        """
        self._graph = graph
        self.node_names = node_names
        assert node_names, "Can't have an empty SEM"
        assert graph.shape[-2:] == (
            len(node_names),
            len(node_names),
        ), "Graph adjacency matrix must have shape [num_nodes, num_nodes] (excluding batch dimensions)"
        super().__init__(event_shape=event_shape, batch_shape=batch_shape)

    @property
    def graph(self) -> torch.Tensor:
        """
        The adjacency matrix representing the graph.
        """
        return self._graph

    @abc.abstractmethod
    def log_prob(self, value: TensorDict) -> torch.Tensor:
        pass

    @abc.abstractmethod
    def do(self, interventions: TensorDict) -> "SEM":
        """Return the SEM associated with the interventions"""

    @abc.abstractmethod
    def noise_to_sample(self, noise: TensorDict) -> TensorDict:
        """
        For a given noise vector, return the corresponding sample values.

        Args:
            noise: Dictionary of tensors representing the noise shape sample_shape + batch_shape + noise_shape
        Return:
            Dictionary of samples from the sem corresponding to the noise, shape sample_shape + batch_shape + event_shape
        """

    @abc.abstractmethod
    def sample_to_noise(self, sample: TensorDict) -> TensorDict:
        """
        For a given sample get the noise vector.

        Args:
            sample: samples from the sem each of shape sample_shape + batch_shape + event_shape
        Return:
            Dictionary of tensors representing the noise shape sample_shape + batch_shape + noise_shape
        """

    @torch.no_grad()
    def sample(self, sample_shape: torch.Size = torch.Size()) -> TensorDict:
        """
        Sample from the SEM

        Grads shall not pass through this method (see `Distribution.sample` vs `Distribution.rsample`).

        Args:
            sample_shape: shape of the returned samples
        Return:
            Dictionary of samples from the sem corresponding to the noise, shape sample_shape + batch_shape + event_shape
        """
        noise_dict = self.sample_noise(sample_shape=sample_shape)
        return self.noise_to_sample(noise=noise_dict)

    @abc.abstractmethod
    @torch.no_grad()
    def sample_noise(self, sample_shape: torch.Size = torch.Size()) -> TensorDict:
        """
        Sample the noise vector for the distribution.

        Grads shall not pass through implementations of this method.

        Args:
            sample_shape: shape of the returned noise samples
        Return:
            Dictionary of samples from the noise distribution for each node, shape sample_shape + batch_shape + event_shape
        """


def ite(
    sem: SEM,
    factual_data: TensorDict,
    intervention_a: TensorDict,
    intervention_b: TensorDict,
    effects: Optional[list[str]] = None,
) -> TensorDict:
    """Calculate ITE of intervention A and B on some factual data for a list of effects.

    Args:
        factual_data: Factual data to abduct the noise from.
        intervention_a: Specification of intervention A.
        intervention_b: Specification of intervention B.
        effects: List of effect variables. Defaults to None.

    Returns:
       TensorDict: Dictionary holding the ITEs for the effect variables for all samples.
    """
    if effects is None:
        effects = list(set(sem.node_names) - set(intervention_a.keys()) - set(intervention_b.keys()))

    base_noise = sem.sample_to_noise(factual_data)
    do_a_cfs = sem.do(interventions=intervention_a).noise_to_sample(base_noise)
    do_b_cfs = sem.do(interventions=intervention_b).noise_to_sample(base_noise)

    return {effect: do_a_cfs[effect] - do_b_cfs[effect] for effect in effects}


def counterfactual(sem: SEM, factual_data: TensorDict, intervention: TensorDict) -> TensorDict:
    """Calculate Counterfactual of an intervention on some factual data.

    Args:
        factual_data: Factual data to abduct the noise from.
        intervention: Specification of intervention.

    Returns:
       TensorDict: Dictionary holding the counterfactual values for all samples.
    """
    # TODO: do we want to average over multiple "sample_to_noise" values for the discrete variables
    # where this is not a 1:1 mapping?
    return sem.do(interventions=intervention).noise_to_sample(sem.sample_to_noise(factual_data))


def ate(
    sem: SEM,
    intervention_a: TensorDict,
    intervention_b: TensorDict,
    effects: Optional[list[str]] = None,
    num_samples: int = 1000,
) -> TensorDict:
    """Calculate the ATE of intervention A and B for a list of effects.

    Args:
        intervention_a: Specification of intervention A.
        intervention_b: Specification of intervention B.
        effects: List of effect variables. Defaults to None.
        num_samples: Number of Monte-Carlo samples to estimate the ATE. Defaults to 1000.

    Returns:
       TensorDict: Dictionary holding the ATEs for the effect variables.
    """
    sample_shape = torch.Size([num_samples])

    if effects is None:
        effects = list(set(sem.node_names) - set(intervention_a.keys()) - set(intervention_b.keys()))

    do_a_samples = sem.do(interventions=intervention_a).sample(sample_shape)
    do_b_samples = sem.do(interventions=intervention_b).sample(sample_shape)

    return TensorDict(
        {effect: do_a_samples[effect].mean(0) - do_b_samples[effect].mean(0) for effect in effects}, batch_size=tuple()
    )



import torch


def calculate_dagness(A: torch.Tensor) -> torch.Tensor:
    """
    Computes the dag penalty for matrix A as trace(exp(A)) - dim.

    Args:
        A (torch.Tensor): Binary adjacency matrix, size (...., input_dim, input_dim).

    Returns:
        (torch.Tensor): Dagness term.
    """
    return torch.einsum("...ii->...", torch.matrix_exp(A)) - A.shape[-1]



from collections import defaultdict
from enum import Enum
from typing import Any, Callable, Iterable, Optional, TypeVar

import torch
from tensordict import TensorDict
from torch import nn

from causica.datasets.causica_dataset_format import VariableTypeEnum
from causica.distributions.noise.bernoulli import BernoulliNoiseModule
from causica.distributions.noise.categorical import CategoricalNoiseModule
from causica.distributions.noise.noise import Noise, NoiseModule
from causica.distributions.noise.spline import SplineNoiseModule
from causica.distributions.noise.univariate_normal import UnivariateNormalNoiseModule

SelectionType = TypeVar("SelectionType", str, Iterable[str])


class JointNoise(Noise[TensorDict]):
    """Represents an independent joint noise distribution of multiple variables types.

    Samples are TensorDicts containining independent variables.
    """

    def __init__(self, independent_noise_dists: dict[str, Noise[torch.Tensor]]):
        shapes = defaultdict[torch.Size, list[str]](list)
        for name, noise_dist in independent_noise_dists.items():
            shapes[noise_dist.batch_shape].append(name)
        if len(shapes) > 1:
            raise ValueError(f"Incompatible batch shapes: {shapes}")

        batch_shape = next(iter(shapes)) if shapes else torch.Size()
        super().__init__(batch_shape=batch_shape, event_shape=torch.Size([len(independent_noise_dists)]))
        self._independent_noise_dists = independent_noise_dists

    def _apply_individually(self, value: TensorDict, func: Callable[[Noise, torch.Tensor], torch.Tensor]) -> TensorDict:
        return TensorDict(
            {name: func(noise_dist, value.get(name)) for name, noise_dist in self._independent_noise_dists.items()},
            batch_size=value.batch_size,
            device=value.device,
        )

    def sample_to_noise(self, samples: TensorDict) -> TensorDict:
        return self._apply_individually(samples, lambda noise_dist, x: noise_dist.sample_to_noise(x))

    def noise_to_sample(self, noise: TensorDict) -> TensorDict:
        return self._apply_individually(noise, lambda noise_dist, x: noise_dist.noise_to_sample(x))

    def sample(self, sample_shape: torch.Size = torch.Size()) -> TensorDict:
        return TensorDict(
            {name: noise_dist.sample(sample_shape) for name, noise_dist in self._independent_noise_dists.items()},
            batch_size=sample_shape + self.batch_shape,
        )

    def log_prob(self, value: TensorDict) -> torch.Tensor:
        """Compute the log probs of the given values.

        Note:
            Produces `log_probs` of the same shape as the inner noise distributions.

        Args:
            value: Values matching the inner independent noise distributions by key. The batch shape of the TensorDict
                   should be set to catch all dims except those associated with the event shape for the expected return
                   shape to behave as specified.

        Returns:
            torch.Tensor: The log probability of the given value, of shape
                           `torch.broadcast_shapes(value.batch_size, self.batch_shape)`.
        """
        log_probs = [noise_dist.log_prob(value[name]) for name, noise_dist in self._independent_noise_dists.items()]
        return torch.sum(torch.stack(log_probs, dim=0), dim=0)

    @property
    def support(self) -> dict[str, Optional[Any]]:
        return {name: noise_dist.support for name, noise_dist in self._independent_noise_dists.items()}

    @property
    def mode(self) -> TensorDict:
        return TensorDict(
            {name: noise_dist.mode for name, noise_dist in self._independent_noise_dists.items()},
            batch_size=self.batch_shape,
        )

    @property
    def mean(self) -> TensorDict:
        return TensorDict(
            {name: noise_dist.mean for name, noise_dist in self._independent_noise_dists.items()},
            batch_size=self.batch_shape,
        )

    def entropy(self) -> torch.Tensor:
        entropies = [noise_dist.entropy() for noise_dist in self._independent_noise_dists.values()]
        return torch.sum(torch.stack(entropies, dim=0), dim=0)


class ContinuousNoiseDist(Enum):
    SPLINE = "spline"
    GAUSSIAN = "gaussian"


def create_noise_modules(
    shapes: dict[str, torch.Size],
    types: dict[str, VariableTypeEnum],
    continuous_noise_dist: ContinuousNoiseDist,
) -> dict[str, NoiseModule[Noise[torch.Tensor]]]:
    """Create noise modules for each item of shapes and types.

    Args:
        shapes: The shape of each distribution. Currently only the last dimension is used.
        types: Names of variables mapping to the variable type `VariableTypeEnum`.
        continuous_noise_dist: The noise module to use for variable types of `VariableTypeEnum.CONTINUOUS`.

    Raises:
        ValueError: If any of the types or the continuous noise distribution is incorrectly specified.

    Returns
        Dict of independent noise modules following the shape and type specifications.
    """
    noise_modules: dict[str, NoiseModule] = {}
    for key, shape in shapes.items():
        size = shape[-1]
        var_type = types[key]

        if var_type == VariableTypeEnum.CATEGORICAL:
            noise_module = CategoricalNoiseModule(size)
        elif var_type == VariableTypeEnum.BINARY:
            noise_module = BernoulliNoiseModule(size)
        elif var_type == VariableTypeEnum.CONTINUOUS:
            if continuous_noise_dist == ContinuousNoiseDist.SPLINE:
                noise_module = SplineNoiseModule(size)
            elif continuous_noise_dist == ContinuousNoiseDist.GAUSSIAN:
                noise_module = UnivariateNormalNoiseModule(size)
            else:
                raise ValueError(f"Invalid continuous noise distribution {continuous_noise_dist}.")
        else:
            raise ValueError(f"Invalid variable type {var_type}")
        noise_modules[key] = noise_module
    return noise_modules


class JointNoiseModule(NoiseModule[JointNoise]):
    """Represents JointNoise with learnable parameters.

    Each noise module is used independently on their corresponding key of sample TensorDicts.
    """

    def __init__(self, independent_noise_modules: dict[str, NoiseModule[Noise[torch.Tensor]]]):
        """
        Args:
            independent_noise_modules: Noise modules to be applied keywise to input TensorDicts. Could e.g. be created
                                       by `create_noise_distributions`.
        """
        super().__init__()
        self.noise_modules = nn.ModuleDict(independent_noise_modules)

    def forward(self, x: Optional[TensorDict] = None) -> JointNoise:
        if x is None:
            noise_distributions = {name: noise_module() for name, noise_module in self.noise_modules.items()}
        else:
            noise_distributions = {name: noise_module(x.get(name)) for name, noise_module in self.noise_modules.items()}
        return JointNoise(independent_noise_dists=noise_distributions)

    def __getitem__(self, selection: Iterable[str]) -> "JointNoiseModule":
        """Return a JointNoiseModule representing the subset of variables specified in selection."""
        selected_independent_noise_modules = {name: self.noise_modules[name] for name in selection}
        # ModuleDict doesn't allow tracking value types, so ignore the type here.
        return JointNoiseModule(selected_independent_noise_modules)  # type: ignore

    def keys(self) -> tuple[str, ...]:
        """The keys for the different noise modules in order."""
        return tuple(self.noise_modules)



"""
Module that provides data normalisation functionality.
"""
import torch
import torch.distributions as td
from tensordict import TensorDict
from torch import nn

from causica.distributions.transforms import JointTransform


class SingleVariableStandardizer(nn.Module):
    """Standardizer module for a single variable, ie a single tensor."""

    def __init__(self, mean: torch.Tensor, std: torch.Tensor):
        """
        Args:
            mean: Mean of the variable
            std: Standard deviation of the variable
        """
        super().__init__()

        if mean.shape != std.shape:
            raise ValueError("mean and std must have the same shape.")

        self.mean: torch.Tensor
        self.std: torch.Tensor

        self.register_buffer("mean", mean)
        self.register_buffer("std", std)

    def __call__(self) -> td.AffineTransform:
        return super().__call__()

    def forward(self) -> td.AffineTransform:
        return td.AffineTransform(loc=-self.mean / self.std, scale=1 / self.std)


class JointStandardizer(nn.Module):
    """Standardizer module for TensorDicts."""

    def __init__(self, means: TensorDict, stds: TensorDict):
        """
        Args:
            means: Means of the variables
            stds: Standard deviations of the variables
        """
        super().__init__()

        if set(means.keys()) != set(stds.keys()):
            raise ValueError("Requires means and stds to have the same keys.")

        if not all(m.shape == s.shape for m, s in zip(means.values(), stds.values())):
            raise ValueError("Requires means and stds to have the same shapes.")

        self.transform_modules = nn.ModuleDict(
            {key: SingleVariableStandardizer(means[key], stds[key]) for key in means.keys()}
        )

    def forward(self) -> JointTransform:
        return JointTransform({key: module() for key, module in self.transform_modules.items()})


def fit_standardizer(data: TensorDict) -> JointStandardizer:
    """Calculate the mean and standard deviation over the first dimension of each variable in the TensorDict and return a standardizer."""
    means = data.apply(lambda x: torch.mean(x, dim=0, keepdim=True), batch_size=(1,))
    # Filter out std == 0
    stds = data.apply(lambda x: torch.std(x, dim=0, keepdim=True), batch_size=(1,)).apply(
        lambda x: torch.where(x == 0, torch.ones_like(x), x)
    )
    return JointStandardizer(means=means, stds=stds)



import torch

from causica.triangular_transformations import unfill_triangular


def adjacency_precision_recall(graph1: torch.Tensor, graph2: torch.Tensor) -> tuple[torch.Tensor, torch.Tensor]:
    """Evaluate the precision and recall of edge existence for two adjacency matrices."""
    vec1 = torch.abs(_to_vector(graph1)) > 0
    vec2 = torch.abs(_to_vector(graph2)) > 0

    correspondence = (vec1 & vec2).sum()

    if (vec1_sum := vec1.sum()) != 0:
        recall = correspondence / vec1_sum
    else:
        recall = torch.tensor(0.0, device=graph1.device)

    if (vec2_sum := vec2.sum()) != 0:
        precision = correspondence / vec2_sum
    else:
        precision = torch.tensor(0.0, device=graph1.device)

    return precision, recall


def adjacency_f1(graph1: torch.Tensor, graph2: torch.Tensor) -> torch.Tensor:
    """Evaluate the f1 score of edge existence for two adjacency matrices."""
    return f1_score(*adjacency_precision_recall(graph1, graph2))


def orientation_precision_recall(graph1: torch.Tensor, graph2: torch.Tensor) -> tuple[torch.Tensor, torch.Tensor]:
    """Evaluate the precision and recall of edge orientation for two adjacency matrices."""
    vec1 = _to_vector(graph1)
    vec2 = _to_vector(graph2)
    non_zero_vec1 = vec1 != 0
    non_zero_vec2 = vec2 != 0

    if (non_zero_vec1_sum := non_zero_vec1.sum()) != 0:
        recall = ((vec1 == vec2) & non_zero_vec1).sum() / non_zero_vec1_sum
    else:
        recall = torch.tensor(0.0, device=graph1.device)

    if (non_zero_vec2_sum := non_zero_vec2.sum()) != 0:
        precision = ((vec1 == vec2) & non_zero_vec2).sum() / non_zero_vec2_sum
    else:
        precision = torch.tensor(0.0, device=graph1.device)

    return precision, recall


def orientation_f1(graph1: torch.Tensor, graph2: torch.Tensor) -> torch.Tensor:
    """Evaluate the f1 score of edge existence for two adjacency matrices."""
    return f1_score(*orientation_precision_recall(graph1, graph2))


def f1_score(precision: torch.Tensor, recall: torch.Tensor) -> torch.Tensor:
    """Calculate f1 score from precision and recall."""
    if torch.abs(denominator := precision + recall) < 1e-8:
        return torch.tensor(0.0, device=precision.device)
    return 2.0 * precision * recall / denominator


def _to_vector(graph: torch.Tensor) -> torch.Tensor:
    """
    Convert an adjacency matrix to a vector of length n(n-1)/2.

    There is a 0 for no edge, -1 or 1 for a single edge and 2 for both edges between a pair of nodes
    """
    lower_tri = unfill_triangular(graph, upper=False)
    upper_tri = unfill_triangular(graph, upper=True)
    diff = upper_tri - lower_tri
    return diff + (1 - torch.abs(diff)) * (upper_tri + lower_tri)



import torch
from tensordict import TensorDict

from causica.distributions import JointNoiseModule
from causica.functional_relationships import FunctionalRelationships, create_do_functional_relationship
from causica.sem.structural_equation_model import SEM


class DistributionParametersSEM(SEM):
    """
    A Structural Equation Model where the functional forward pass generates the parameters of the distributions to be sampled.

    This is more general than the simple additive case.
    """

    arg_constraints: dict = {}

    def __init__(
        self,
        graph: torch.Tensor,
        noise_dist: JointNoiseModule,
        func: FunctionalRelationships,
    ):
        """
        Args:
            graph: Adjacency Matrix
            noise_dist: A dictionary of callables for generating distributions from predictions
            func: The functional relationship that can map from the tensor representation to the to itself.
        """
        super().__init__(graph=graph, node_names=noise_dist.keys())
        self.noise_dist = noise_dist
        self.func = func

    def log_prob(self, value: TensorDict) -> torch.Tensor:
        """
        Compute the log prob of the observations

        Args:
            value: a dictionary of `sample_shape + batch_shape + [variable_dim]` shaped tensors
        Return:
            A tensor of shape `sample_shape + batch_shape` of the log probability of each event
        """
        return self.noise_dist(self.func(value, self.graph)).log_prob(value)

    def noise_to_sample(self, noise: TensorDict) -> TensorDict:
        """
        For a given noise vector, return the corresponding sample values.

        Args:
            noise: Dictionary of tensors representing the noise shape
            `sample_shape + batch_shape + noise_shape`
        Return:
            Dictionary of samples from the sem corresponding to the noise, shape
            `sample_shape + batch_shape + event_shape`
        """
        x = noise.clone().zero_()
        for _ in self.node_names:
            x.update_(self.noise_dist(self.func(x, self.graph)).noise_to_sample(noise))
        return x

    def sample_to_noise(self, sample: TensorDict) -> TensorDict:
        """
        For a given sample get the noise vector.

        Args:
            sample: samples from the sem each of shape `sample_shape + batch_shape + event_shape`
        Return:
            Dictionary of tensors representing the noise shape `sample_shape + batch_shape + noise_shape`
        """
        return self.noise_dist(self.func(sample, self.graph)).sample_to_noise(sample)

    @torch.no_grad()
    def sample_noise(self, sample_shape: torch.Size = torch.Size()) -> TensorDict:
        """
        Sample the noise vector for the distribution.

        Args:
            sample_shape: shape of the returned noise samples
        Return:
            Dictionary of samples from the noise distribution for each node, shape
            `sample_shape + batch_shape + event_shape`.
        """
        return self.noise_dist().sample(sample_shape)

    def do(self, interventions: TensorDict) -> "DistributionParametersSEM":
        """Return the SEM associated with the interventions"""
        do_func, do_graphs = create_do_functional_relationship(
            interventions=interventions, func=self.func, graph=self.graph
        )
        intervened_node_names = set(interventions.keys())
        unintervened_node_names = [name for name in self.node_names if name not in intervened_node_names]  # keep order
        return DistributionParametersSEM(
            graph=do_graphs,
            noise_dist=self.noise_dist[unintervened_node_names],
            func=do_func,
        )



from pathlib import Path
from tempfile import TemporaryDirectory
from typing import Any

import mlflow
import pytorch_lightning as pl
import torch
from pytorch_lightning.cli import LightningArgumentParser, Namespace, SaveConfigCallback
from pytorch_lightning.trainer.states import TrainerFn
from pytorch_lightning.utilities.types import STEP_OUTPUT

from causica.training.auglag import AugLagLossCalculator, AugLagLR


class AuglagLRCallback(pl.Callback):
    """Wrapper Class to make the Auglag Learning Rate Scheduler compatible with Pytorch Lightning"""

    def __init__(self, scheduler: AugLagLR):
        """
        Args:
            scheduler: The auglag learning rate scheduler to wrap
        """
        self.scheduler = scheduler

    def on_train_batch_end(
        self, trainer: pl.Trainer, pl_module: pl.LightningModule, outputs: STEP_OUTPUT, batch: Any, batch_idx: int
    ) -> None:
        _ = trainer
        _ = batch
        _ = batch_idx
        assert isinstance(outputs, dict)
        optimizer = pl_module.optimizers()
        assert isinstance(optimizer, torch.optim.Optimizer)
        auglag_loss: AugLagLossCalculator = pl_module.auglag_loss  # type: ignore

        is_converged = self.scheduler.step(
            optimizer=optimizer,
            loss=auglag_loss,
            loss_value=outputs["loss"].item(),
            lagrangian_penalty=outputs["constraint"].item(),
        )

        auglag_logging_attributes = [
            "num_lr_updates",
            "outer_opt_counter",
            "step_counter",
            "outer_below_penalty_tol",
            "outer_max_rho",
            "last_best_step",
            "last_lr_update_step",
        ]
        pl_module.log_dict({f"auglag/{k}": getattr(self.scheduler, k) for k in auglag_logging_attributes}, on_step=True)

        # Notify trainer to stop if the auglag algorithm has converged
        if is_converged:
            trainer.should_stop = True


class MLFlowSaveConfigCallback(SaveConfigCallback):
    """Logs the config using MLFlow if there is an active run, otherwise saves locally as the superclass."""

    def __init__(
        self,
        parser: LightningArgumentParser,
        config: Namespace,
        config_filename: str = "config.yaml",
        multifile: bool = False,
    ) -> None:
        super().__init__(
            parser=parser, config=config, config_filename=config_filename, overwrite=True, multifile=multifile
        )

    def setup(self, trainer: pl.Trainer, pl_module: pl.LightningModule, stage: TrainerFn) -> None:  # type: ignore
        # Save the file on rank 0
        if trainer.is_global_zero and stage == TrainerFn.FITTING:
            with TemporaryDirectory() as tmpdir:
                temporary_config_path = str(Path(tmpdir) / self.config_filename)
                self.parser.save(
                    self.config,
                    temporary_config_path,
                    skip_none=False,
                    overwrite=self.overwrite,
                    multifile=self.multifile,
                )
                # AzureML throws a raw Exception if the artifact already exists, so we check the error message
                try:
                    mlflow.log_artifact(temporary_config_path)
                # pylint: disable=broad-exception-caught
                except Exception as e:
                    if "Resource Conflict" not in str(e.args):
                        raise e



import math
from collections import defaultdict
from typing import Iterable

import torch
from tensordict import TensorDict

from causica.datasets.causica_dataset_format import CounterfactualWithEffects, InterventionWithEffects
from causica.sem.structural_equation_model import SEM, ate, ite


def eval_intervention_likelihoods(sems: list[SEM], intervention_with_effects: InterventionWithEffects) -> torch.Tensor:
    """
    Calculate the average log-prob of interventional data.

    Specifically we calculate 𝔼_sample[log(𝔼_G[p(sample | G)])]

    Args:
        sems: An iterable of SEMS to evaluate the interventional log prob of
        interventions: True interventional data to use for evaluation.

    Returns:
        Log-likelihood of the interventional data for each interventional datapoint
    """
    total_log_sum_exp_per_int = []  # this will end up being length number of interventions
    for intervention in intervention_with_effects[:2]:
        inner_log_probs = [
            sem.do(interventions=intervention.intervention_values).log_prob(intervention.intervention_data)
            for sem in sems
        ]
        # calculate log(𝔼_G[p(sample | G)]) for each sample
        log_prob = list_logsumexp(inner_log_probs) - math.log(len(sems))  # batch_size
        assert len(log_prob.shape) == 1
        total_log_sum_exp_per_int.append(log_prob)

    # log(𝔼_G[p(sample | G)]) for each sample in both interventions, shape 2 * batch_size
    return torch.cat(total_log_sum_exp_per_int, dim=-1)


def eval_ate_rmse(
    sems: Iterable[SEM], intervention: InterventionWithEffects, samples_per_graph: int = 1000
) -> TensorDict:
    """Evaluate the ATEs of a model

    Args:
        sems: An iterable of structural equation models to evaluate the ATE RMSE of
        intervention: True interventional data to use for evaluation.
        samples_per_graph: Number of samples to draw per graph to calculate the ATE.

    Returns:
        Dict of the RMSE of the ATE for each node we're interested in
    """

    intervention_a, intervention_b, effects = intervention

    # each node has shape [batch_size, node_dim]
    true_ates = {
        effect: intervention_a.intervention_data[effect].mean() - intervention_b.intervention_data[effect].mean()
        for effect in effects
    }

    # generate samples from the intervened distribution and the base distribution
    ates_per_graph: dict[str, list[torch.Tensor]] = defaultdict(list)
    for sem in sems:
        graph_ates = ate(
            sem, intervention_a.intervention_values, intervention_b.intervention_values, effects, samples_per_graph
        )
        for key in effects:
            ates_per_graph[key].append(graph_ates[key].detach())

    # each node has shape [node_dim]
    generated_ates = {k: list_mean(v) for k, v in ates_per_graph.items()}
    return TensorDict(
        {key: torch.sqrt(torch.mean(torch.sum((generated_ates[key] - true_ates[key]) ** 2, -1))) for key in effects},
        batch_size=torch.Size([]),
    )


def eval_ite_rmse(sems: Iterable[SEM], counterfactual_data: CounterfactualWithEffects) -> TensorDict:
    """Evaluate the ITEs of a model.

    Args:
        sems: An iterable of structural equation models to evaluate the ITE RMSE of
        counterfactual_data: Data of true counterfactuals to use for evaluation.

    Returns:
        Dict of RMSEs for each effect variable we're interested in
    """
    intervention_a, intervention_b, effects = counterfactual_data
    if intervention_b is None:
        raise ValueError("ITE evaluation must have reference counterfactuals")
    for key, a_val in intervention_a.factual_data.items():
        b_val = intervention_b.factual_data[key]
        assert torch.allclose(a_val, b_val), "Base data must be the same for ITEs"
    # each node has shape [batch_size, node_dim]
    true_ites = {
        effect: intervention_a.counterfactual_data[effect] - intervention_b.counterfactual_data[effect]
        for effect in effects
    }

    # generate samples from the intervened distribution and the base distribution
    per_graph_ites: dict[str, list[torch.Tensor]] = defaultdict(list)
    for sem in sems:
        sem_ites = ite(
            sem,
            intervention_a.factual_data,
            intervention_a.intervention_values,
            intervention_b.intervention_values,
            effects,
        )

        for key in effects:
            per_graph_ites[key].append(sem_ites[key].detach())

    # average the treatment value over all graphs, each node has shape [batch_size, node_dim]
    generated_ites = {k: list_mean(v) for k, v in per_graph_ites.items()}
    return TensorDict(
        {key: torch.sqrt(torch.mean(torch.sum((generated_ites[key] - true_ites[key]) ** 2, -1))) for key in effects},
        batch_size=torch.Size(),
    )


def list_mean(list_of_tensors: list[torch.Tensor]) -> torch.Tensor:
    """Take the mean of a list of torch tensors, they must all have the same shape"""
    return torch.stack(list_of_tensors, dim=0).mean(dim=0)


def list_logsumexp(list_of_tensors: list[torch.Tensor]) -> torch.Tensor:
    """Take the logsumexp of a list of torch tensors, they must all have the same shape"""
    return torch.logsumexp(torch.stack(list_of_tensors, dim=0), dim=0)



from typing import Optional

import torch
import torch.distributions as td
from torch import nn

from causica.distributions.noise.noise import IndependentNoise, Noise, NoiseModule


class UnivariateNormalNoise(td.Normal, Noise[torch.Tensor]):
    def sample_to_noise(self, samples: torch.Tensor) -> torch.Tensor:
        """
        Transform from the sample observations to corresponding noise variables.

        Args:
            samples: Tensor of shape sample_shape + batch_shape + event_shape
        Returns:
            The generated samples with shape sample_shape + batch_shape + event_shape
        """
        return samples - self.loc

    def noise_to_sample(self, noise: torch.Tensor) -> torch.Tensor:
        """
        Generate samples using the given exogenous noise.

        Args:
            noise: noise variable with shape sample_shape + batch_shape.
        Returns:
            The generated samples with shape sample_shape + batch_shape + event_shape
        """
        return noise + self.loc


class UnivariateNormalNoiseModule(NoiseModule[UnivariateNormalNoise]):
    """Represents a UnivariateNormalNoise with learnable parameters for independent variables."""

    def __init__(self, dim: int, init_log_scale: float = 0.0):
        """
        Args:
            dim: Number of dimensions for the NormalNoise.
        """
        super().__init__()
        self.log_scale = nn.Parameter(torch.full(torch.Size([dim]), init_log_scale))

    def forward(self, x: Optional[torch.Tensor] = None) -> IndependentNoise[UnivariateNormalNoise]:
        if x is None:
            x = torch.zeros_like(self.log_scale)
        return IndependentNoise(UnivariateNormalNoise(loc=x, scale=torch.exp(self.log_scale)), 1)



from enum import Enum

import torch


class VariableTypeEnum(Enum):
    CONTINUOUS = "continuous"
    CATEGORICAL = "categorical"
    BINARY = "binary"


DTYPE_MAP = {
    VariableTypeEnum.CONTINUOUS: torch.float32,
    VariableTypeEnum.CATEGORICAL: torch.int32,
    VariableTypeEnum.BINARY: torch.int32,
}



import abc

import pytorch_lightning as pl
import torch
from tensordict import TensorDict

from causica.datasets.variable_types import VariableTypeEnum


class DECIDataModule(pl.LightningDataModule, abc.ABC):
    """An Abstract Data Module containing the methods required by a `DECIModule`."""

    @property
    @abc.abstractmethod
    def dataset_name(self) -> str:
        """The name of this dataset"""

    @property
    @abc.abstractmethod
    def dataset_train(self) -> TensorDict:
        """The training dataset"""

    @property
    @abc.abstractmethod
    def dataset_test(self) -> TensorDict:
        """The testing dataset"""

    @property
    @abc.abstractmethod
    def variable_shapes(self) -> dict[str, torch.Size]:
        """Get the shape of each variable in the dataset."""

    @property
    @abc.abstractmethod
    def variable_types(self) -> dict[str, VariableTypeEnum]:
        """Get the type of each variable in the dataset."""

    @property
    @abc.abstractmethod
    def column_names(self) -> dict[str, list[str]]:
        """Get a map of the node names and the corresponding columns of the original dataset."""



from causica.distributions.adjacency.adjacency_distributions import AdjacencyDistribution
from causica.distributions.adjacency.constrained_adjacency_distributions import (
    ConstrainedAdjacency,
    ConstrainedAdjacencyDistribution,
)
from causica.distributions.adjacency.directed_acyclic import ErdosRenyiDAGDistribution
from causica.distributions.adjacency.enco import ENCOAdjacencyDistribution, ENCOAdjacencyDistributionModule
from causica.distributions.adjacency.gibbs_dag_prior import ExpertGraphContainer, GibbsDAGPrior
from causica.distributions.adjacency.three_way import ThreeWayAdjacencyDistribution



from dataclasses import dataclass, field

import torch
from tensordict import TensorDict


@dataclass
class InterventionData:
    """
    Dataclass to hold the data associated with an intervention

    This represents one intervention and many samples from the intervened distribution

    The class also stores `sampled_nodes`, i.e. the ones that are neither intervened or conditioned on

    Args:
        intervention_data: A `TensorDict` with all the nodes (including intervened/conditioned)
        intervention_values: A dictionary of node names to 1D numpy arrays of the intervened values
        condition_values: A dictionary of node names to 1D numpy arrays of the conditioned values
    """

    intervention_data: TensorDict
    intervention_values: TensorDict
    condition_values: TensorDict
    sampled_nodes: set[str] = field(init=False)  # the nodes that are neither conditioned nor sampled

    def __post_init__(self):
        assert self.intervention_values.batch_size == torch.Size()
        assert self.condition_values.batch_size == torch.Size()

        self.sampled_nodes = (
            set(self.intervention_data.keys())
            - set(self.intervention_values.keys())
            - set(self.condition_values.keys())
        )


@dataclass
class CounterfactualData:
    """
    Dataclass to hold the data associated with a counterfactual

    This represents one intervention and reference and many samples from the intervened and reference
    distributions

    The class also stores `sampled_nodes`, i.e. the ones that are neither intervened or conditioned on

    Args:
        counterfactual_data: A `TensorDict` with all of the node values (including intervened) of counterfactual data
        factual_data: A `TensorDict` with all of the node values of the base observations used for the counterfactuals data.
            This refers to the observations in "What would have happened (CFs) if I would have done (intervention) given
            I observed (base observation).
        intervention_values: A dictionary of node names to 1D numpy arrays of the intervened values
    """

    counterfactual_data: TensorDict
    intervention_values: TensorDict
    factual_data: TensorDict
    sampled_nodes: set[str] = field(init=False)

    def __post_init__(self):
        assert list(self.counterfactual_data.keys()) == list(self.factual_data.keys())
        assert self.counterfactual_data.batch_size == self.factual_data.batch_size
        assert self.intervention_values.batch_size == torch.Size()

        self.sampled_nodes = set(self.counterfactual_data.keys()) - set(self.intervention_values.keys())



import abc
from typing import Any

import torch
from tensordict import TensorDict


class FunctionalRelationships(abc.ABC, torch.nn.Module):
    def __init__(self, variables: dict[str, torch.Size]) -> None:
        """_summary_

        Args:
            variables: Dict of node shapes (how many dimensions a variable has)
                Order corresponds to the order in graph(s).
        """
        super().__init__()
        self.num_nodes = len(variables)
        self.variables = variables
        self.output_shape = sum(variable.numel() for variable in variables.values())

        self.variable_masks = {}
        last_idx = 0
        for name, shape in variables.items():
            mask = torch.zeros(self.output_shape, dtype=torch.bool)
            mask[last_idx : last_idx + shape.numel()] = True

            self.variable_masks[name] = mask
            last_idx += shape.numel()

    def set_extra_state(self, state: dict[str, Any]):
        self.num_nodes = state.pop("num_nodes")
        self.variables = state.pop("variables")
        self.output_shape = state.pop("output_shape")

    def get_extra_state(self) -> dict[str, Any]:
        return {
            "num_nodes": self.num_nodes,
            "variables": self.variables,
            "output_shape": self.output_shape,
        }

    @abc.abstractmethod
    def forward(self, samples: TensorDict, graphs: torch.Tensor) -> TensorDict:
        """Calculates the predictions of the children from parents.

        Args:
            samples: dictionary of variable samples of shape sample_shape + [node shape]
            graphs: tensor of shape batch_shape + [nodes, nodes]

        Returns:
            Dictionary of torch.Tensors of shape sample_shape + batch_shape + [node shape]
        """


def sample_dict_to_tensor(sample_dict: TensorDict, variable_masks: dict[str, torch.Tensor]) -> torch.Tensor:
    """Converts a sample dictionary to a tensor."""
    return torch.cat([sample_dict[name] for name in variable_masks.keys()], dim=-1)


def tensor_to_sample_dict(sample_tensor: torch.Tensor, variable_masks: dict[str, torch.Tensor]) -> TensorDict:
    """Converts a tensor to a sample dictionary."""
    return TensorDict(
        {name: sample_tensor[..., mask] for name, mask in variable_masks.items()}, batch_size=sample_tensor.shape[:-1]
    )



import os

from pytorch_lightning.callbacks import ModelCheckpoint
from pytorch_lightning.cli import LightningCLI


class LightningCLIWithDefaults(LightningCLI):
    default_logger = {
        "class_path": "pytorch_lightning.loggers.MLFlowLogger",
        "init_args": {"run_id": os.environ.get("AZUREML_RUN_ID", None)},
    }

    def add_arguments_to_parser(self, parser):
        parser.add_lightning_class_args(ModelCheckpoint, "best_checkpoint_callback")
        parser.add_lightning_class_args(ModelCheckpoint, "last_checkpoint_callback")
        parser.link_arguments("best_checkpoint_callback.dirpath", "last_checkpoint_callback.dirpath")

        parser.set_defaults({"trainer.logger": self.default_logger})



import numpy as np
import torch


def fill_triangular(vec: torch.Tensor, upper: bool = False) -> torch.Tensor:
    """
    Args:
        vec: A tensor of shape (..., n(n-1)/2)
        upper: whether to fill the upper or lower triangle
    Returns:
        An array of shape (..., n, n), where the strictly upper (lower) triangle is filled from vec
        with zeros elsewhere
    """
    num_nodes = num_lower_tri_elements_to_n(vec.shape[-1])
    idxs = torch.triu_indices(num_nodes, num_nodes, offset=1, device=vec.device)
    output = torch.zeros(vec.shape[:-1] + (num_nodes, num_nodes), device=vec.device)
    output[..., idxs[0, :], idxs[1, :]] = vec
    return output if upper else output.transpose(-1, -2)


def unfill_triangular(mat: torch.Tensor, upper: bool = False) -> torch.Tensor:
    """
    Fill a vector of length n(n-1)/2 with elements from the strictly upper(lower) triangle.

    Args:
        mat: A tensor of shape (..., n, n)
        upper: whether to fill from the upper triangle
    Returns:
        A vector of shape (..., n(n-1)/2), filled from the upper triangle
    """
    num_nodes = mat.shape[-1]
    idxs = torch.triu_indices(num_nodes, num_nodes, offset=1, device=mat.device)
    matrix = mat if upper else mat.transpose(-2, -1)
    return matrix[..., idxs[0, :], idxs[1, :]]


def num_lower_tri_elements_to_n(x: int) -> int:
    """
    Calculate the size of the matrix from the number of strictly lower triangular elements.

    We have x = n(n - 1) / 2 for some n
    n² - n - 2x = 0
    so n = (1 + √(1 + 8x)) / 2
    """
    val = int(np.sqrt(1 + 8 * x) + 1) // 2
    if val * (val - 1) != 2 * x:
        raise ValueError("Invalid number of lower triangular elements")
    return val



from typing import Generic, TypeVar

from torch import distributions as td
from torch import nn

DistributionType_co = TypeVar("DistributionType_co", bound=td.Distribution, covariant=True)


class DistributionModule(Generic[DistributionType_co], nn.Module):
    """Baseclass for modules returning distributions.

    Useful e.g. to create variational approximations of distributions.

    Subclasses are expected to implement a `forward` method that returns a concrete `td.Distribution` and should usually
    inherit from a conrete version of this class, i.e. `DistributionModule[<td.Distribution subclass>]`.
    """

    def __call__(self, *args, **kwargs) -> DistributionType_co:
        """Return a td.Distribution."""
        return super().__call__(*args, **kwargs)



import abc
from typing import Optional

import torch
import torch.distributions as td


class AdjacencyDistribution(td.Distribution, abc.ABC):
    """
    Probability distributions over binary adjacency matrices for graphs.

    NOTE: Because we want to differentiate through samples (which are binary matrices),
    we add a `relaxed_sample` method. This method is still expected to return valid samples (binary matrices).

    Since the statistics of the relaxed distribution can be difficult to calculate,
    we choose to implement `relaxed_sample` as the relaxed distribution but report other statistics for
    the underlying distribution, as well as providing the usual `sample` method for it.

    Clearly this approximation could be invalid when the relaxed distribution is very different from
    the underlying one (at high temperatures).
    """

    support = td.constraints.independent(td.constraints.boolean, 1)

    def __init__(self, num_nodes: int, validate_args: Optional[bool] = None):
        assert num_nodes > 0, "Number of nodes in the graph must be greater than 0"
        self.num_nodes = num_nodes
        event_shape = torch.Size((num_nodes, num_nodes))

        super().__init__(event_shape=event_shape, validate_args=validate_args)

    @abc.abstractmethod
    def relaxed_sample(self, sample_shape: torch.Size = torch.Size(), temperature: float = 0.0) -> torch.Tensor:
        """
        Sample a binary adjacency matrix from the relaxed distribution (see NOTE in the class docstring).

        Args:
            sample_shape: the shape of the samples to return
            temperature: The temperature of the relaxed distribution
        Returns:
            A tensor of shape sample_shape + batch_shape + (num_nodes, num_nodes)
        """
        raise NotImplementedError

    @abc.abstractmethod
    def sample(self, sample_shape: torch.Size = torch.Size()) -> torch.Tensor:
        """
        Sample a binary adjacency matrix from the underlying distribution.

        Args:
            sample_shape: the shape of the samples to return
        Returns:
            A tensor of shape sample_shape + batch_shape + (num_nodes, num_nodes)
        """
        raise NotImplementedError

    @abc.abstractmethod
    def entropy(self) -> torch.Tensor:
        """
        Return the entropy of the underlying distribution.

        Returns:
            A tensor of shape batch_shape, with the entropy of the distribution
        """

        raise NotImplementedError

    @property
    @abc.abstractmethod
    def mean(self) -> torch.Tensor:
        """
        Return the mean of the underlying distribution.

        This will be a matrix with all entries in the interval [0, 1].

        Returns:
            A tensor of shape batch_shape + (num_nodes, num_nodes)
        """
        raise NotImplementedError

    @property
    @abc.abstractmethod
    def mode(self) -> torch.Tensor:
        """
        Return the mode of the underlying distribution.

        This will be an adjacency matrix.

        Returns:
            A tensor of shape batch_shape + (num_nodes, num_nodes)
        """
        raise NotImplementedError

    @abc.abstractmethod
    def log_prob(self, value: torch.Tensor) -> torch.Tensor:
        """
        Get the log probability of each tensor from the sample space

        Args:
            value: a binary matrix of shape value_shape + batch_shape + (n, n)
        Returns:
            A tensor of shape value_shape + batch_shape, with the log probabilities of each tensor in the batch.
        """
        raise NotImplementedError



from typing import Optional

import torch
import torch.distributions as td


class ExpertGraphContainer(torch.nn.Module):
    def __init__(self, dag: torch.Tensor, mask: torch.Tensor, confidence: float, scale: float) -> None:
        """Container holding an "experts" prior belief about the underlying causal DAG.

        Arguments:
            dag: The binary adjacency matrix representing domain knowledge in the form of a DAG. Corresponds to `mask`.
            mask: A binary mask indicating whether or not the corresponding edge of the `dag` has information.
            confidence: A value in the interval (0, 1] indicating the confidence of the existence of the edge
            scale: Scaling factor for expert graph loss
        """
        super().__init__()

        self.dag = torch.nn.Parameter(dag, requires_grad=False)
        self.mask = torch.nn.Parameter(mask, requires_grad=False)
        self.confidence = confidence
        self.scale = scale


class GibbsDAGPrior(td.Distribution):
    """
    Represents a prior distribution over adjacency matrices.

    The prior distribution consists of two terms:
        A sparsity term
        A Expert Graph term that represents some known prior belief about the graph.

    Each term has an associated parameter (lambda)
    """

    arg_constraints: dict = {}

    def __init__(
        self,
        num_nodes: int,
        sparsity_lambda: float,
        expert_graph_container: Optional[ExpertGraphContainer] = None,
        **kwargs
    ) -> None:
        """
        Args:
            num_nodes (int): Number of nodes in the graph
            sparsity_lambda (float): Coefficient of sparsity term
            dagness_alpha (torch.Tensor): Coefficient of dagness term
            dagness_rho (torch.Tensor): Coefficient of squared dagness term
            expert_graph_container (ExpertGraphContainer): Dataclass containing prior belief about the real graph
        """

        super().__init__(torch.Size(), event_shape=torch.Size((num_nodes, num_nodes)), **kwargs)

        self._num_nodes = num_nodes
        self._expert_graph_container = expert_graph_container
        self._sparsity_lambda = sparsity_lambda

    def get_sparsity_term(self, A: torch.Tensor) -> torch.Tensor:
        """
        A term that encourages sparsity (see https://arxiv.org/pdf/2106.07635.pdf).
        The term is small when A is sparse.

        Args:
            A (torch.Tensor): Adjacency matrix of shape (input_dim, input_dim).

        Returns:
            Sparsity term.
        """
        return A.abs().sum()

    def get_expert_graph_term(self, A: torch.Tensor) -> torch.Tensor:
        """
        A term that encourages A to be close to given expert graph.

        Args:
            A (torch.Tensor): Adjacency matrix of shape (input_dim, input_dim).

        Returns:
            (torch.Tensor): Expert graph term.

        """
        assert isinstance(self._expert_graph_container, ExpertGraphContainer)
        return (
            (
                self._expert_graph_container.mask
                * (A - self._expert_graph_container.confidence * self._expert_graph_container.dag)
            )
            .abs()
            .sum()
        )

    def log_prob(self, value: torch.Tensor) -> torch.Tensor:
        """
        Calculates the (un-normalized) log probability of adjacency matrix `value`
        under the distribution given by this instance.

        Args:
            value (torch.Tensor): Adjacency matrix of shape (input_dim, input_dim).

        Returns:
            (torch.Tensor): The un-normalized log probability of `value`.

        """
        assert value.shape[-2:] == (self._num_nodes, self._num_nodes)

        log_prob = -self._sparsity_lambda * self.get_sparsity_term(value)

        if self._expert_graph_container is not None:
            log_prob -= self._expert_graph_container.scale * self.get_expert_graph_term(value)

        return log_prob



from .do_functional_relationships import DoFunctionalRelationships, create_do_functional_relationship
from .functional_relationships import FunctionalRelationships, sample_dict_to_tensor, tensor_to_sample_dict
from .icgnn import ICGNN
from .linear_functional_relationships import LinearFunctionalRelationships



from typing import Optional

import torch
import torch.distributions as td
import torch.nn.functional as F

from causica.distributions.adjacency.adjacency_distributions import AdjacencyDistribution
from causica.triangular_transformations import fill_triangular, num_lower_tri_elements_to_n, unfill_triangular


class ThreeWayAdjacencyDistribution(AdjacencyDistribution):
    """
    For each pair of nodes xᵢ and xⱼ where i < j, sample a three way categorical Cᵢⱼ.
        Cᵢⱼ = 0, represents the edge xᵢ -> xⱼ,
        Cᵢⱼ = 1, represents the edge xᵢ <- xⱼ,
        Cᵢⱼ = 2, there is no edge between these nodes.

    We store the logits of C as a tensor of shape (n(n-1)/2, 3), so we have a three way
    categorical distribution for each edge in the strictly lower triangle.
    """

    arg_constraints = {"logits": td.constraints.real}

    def __init__(self, logits: torch.Tensor, validate_args: Optional[bool] = None):
        """
        Args:
            logits: An array of size (..., n(n-1)/2, 3), representing logit of i->j, j->i, no edge respectively.
            validate_args: Whether to validate the arguments. Passed to the superclass
        """
        num_nodes = num_lower_tri_elements_to_n(logits.shape[-2])

        if validate_args:
            assert len(logits.shape) >= 2, "Logits must be at least dimensional"
            assert logits.shape[-2:] == ((num_nodes * (num_nodes - 1)) // 2, 3), "Invalid logits shape"

        self.logits = logits

        super().__init__(num_nodes, validate_args=validate_args)

    @staticmethod
    def base_dist(logits: torch.Tensor) -> td.Distribution:
        """The underlying distribution is a vector of one hot categorical variables."""
        return td.Independent(td.OneHotCategorical(logits=logits, validate_args=False), 1)

    def relaxed_sample(self, sample_shape: torch.Size = torch.Size(), temperature: float = 0.0) -> torch.Tensor:
        """Use the Gumbel Softmax trick to relax the distribution."""
        shape = sample_shape + self.logits.shape
        samples = F.gumbel_softmax(
            self.logits.expand(*shape), tau=temperature, hard=True, dim=-1
        )  # (..., n(n-1)/2, 3) binary
        return _triangular_vec_to_matrix(samples)

    def sample(self, sample_shape: torch.Size = torch.Size()) -> torch.Tensor:
        """
        Sample a binary adjacency matrix from the underlying distribution.

        Args:
            sample_shape: the shape of the samples to return
        Returns:
            A tensor of shape sample_shape + batch_shape + (num_nodes, num_nodes)
        """
        return _triangular_vec_to_matrix(self.base_dist(self.logits).sample(sample_shape))

    def entropy(self) -> torch.Tensor:
        """
        Return the entropy of the underlying distribution.

        Returns:
            A tensor of shape batch_shape, with the entropy of the distribution
        """
        return self.base_dist(self.logits).entropy()

    @property
    def mean(self) -> torch.Tensor:
        """
        Return the mean of the underlying independent Bernoulli distribution.

        This will be a matrix with all entries in the interval [0, 1].

        Returns:
            A tensor of shape batch_shape + (num_nodes, num_nodes)
        """
        return _triangular_vec_to_matrix(self.base_dist(self.logits).mean)

    @property
    def mode(self) -> torch.Tensor:
        """
        Return the mode of the underlying distribution.

        This will be an adjacency matrix.

        Returns:
            A tensor of shape batch_shape + (num_nodes, num_nodes)
        """
        # bernoulli mode can be nan for very small logits, favour sparseness and set to 0
        return _triangular_vec_to_matrix(torch.nan_to_num(self.base_dist(self.logits).mode, 0.0))

    def log_prob(self, value: torch.Tensor) -> torch.Tensor:
        """
        Get the log probability of each tensor from the sample space

        Args:
            value: a binary matrix of shape value_shape + batch_shape + (n, n)
        Returns:
            A tensor of shape value_shape + batch_shape, with the log probabilities of each tensor in the batch.
        """
        low_tri = unfill_triangular(value, upper=False)
        upp_tri = unfill_triangular(value, upper=True)
        missing = (1 - low_tri) * (1 - upp_tri)
        return self.base_dist(self.logits).log_prob(torch.stack([low_tri, upp_tri, missing], dim=-1))


def _triangular_vec_to_matrix(vec: torch.Tensor) -> torch.Tensor:
    """
    Args:
        vec: A tensor of shape (..., n(n-1)/2, 3)
    Returns:
        A matrix of shape (..., n, n), where the strictly lower triangle is filled from vec[..., 0],
        the strictly upper triangle is filled from vec[..., 1], and zeros on the main diagonal
    """
    return fill_triangular(vec[..., 0], upper=False) + fill_triangular(vec[..., 1], upper=True)



"""
Implementation of the Piecewise Rational Quadratic Transform

This is pretty much a copy-paste of
    https://github.com/tonyduan/normalizing-flows/blob/master/nf/utils.py

We should consider using the Pyro implementation.
"""

import torch
import torch.distributions as td

from causica.distributions.noise.spline.bayesiains_nsf_rqs import unconstrained_rational_quadratic_spline

DEFAULT_MIN_BIN_WIDTH = 1e-3
DEFAULT_MIN_BIN_HEIGHT = 1e-3
DEFAULT_MIN_DERIVATIVE = 1e-3


class PiecewiseRationalQuadraticTransform(td.Transform):
    """
    Layer that implements a spline-cdf (https://arxiv.org/abs/1906.04032) transformation.

    All dimensions of x are treated as independent, no coupling is used. This is needed
    to ensure invertibility in our additive noise SEM.

    This is pretty much a copy-paste of
        https://github.com/tonyduan/normalizing-flows/blob/master/nf/utils.py
    """

    bijective = True

    def __init__(self, knot_locations: torch.Tensor, derivatives: torch.Tensor, tail_bound: float = 3.0):
        """
        Args:
            knot_locations: the x, y points of the knots,  shape [dim, num_bins, 2]
            derivatives: the derivatives at the knots, shape [dim, num_bins - 1]
            tail_bound: distance of edgemost bins relative to 0,
        """
        super().__init__()
        assert knot_locations.ndim == 3, "Only two dimensional params are supported"
        assert knot_locations.shape[-1] == 2, "Knot locations are 2-d"
        self.dim, self.num_bins, *_ = knot_locations.shape

        assert derivatives.shape == (self.dim, self.num_bins - 1)

        self.tail_bound = tail_bound
        self._event_dim = 0

        self.knot_locations = knot_locations
        self.derivatives = derivatives

    @property
    def event_dim(self):
        return self._event_dim

    def _piecewise_cdf(self, inputs: torch.Tensor, inverse: bool = False) -> tuple[torch.Tensor, torch.Tensor]:
        """
        Evaluate the Cumulative Density function at `inputs`

        Args:
            inputs: the positions at which to evaluate the cdf shape batch_shape + (input_dim)
            inverse: whether this is forwards or backwards transform
        Returns:
            input_evaluations and absolute log determinants, a tuple of tensors of shape batch_shape + (input_dim)
        """
        assert len(inputs.shape) == 2  # TODO(JJ) accept 1d inputs
        batch_shape = inputs.shape[:-1]
        # shape batch_shape + (dim, 3 * num_bins - 1)
        expanded_knot_locations = self.knot_locations[None, ...].expand(*batch_shape, -1, -1, -1)
        expanded_derivatives = self.derivatives[None, ...].expand(*batch_shape, -1, -1)

        return unconstrained_rational_quadratic_spline(
            inputs=inputs,
            unnormalized_widths=expanded_knot_locations[..., 0],
            unnormalized_heights=expanded_knot_locations[..., 1],
            unnormalized_derivatives=expanded_derivatives,
            inverse=inverse,
            tail_bound=self.tail_bound,
        )

    @td.constraints.dependent_property(is_discrete=False)
    def domain(self):
        if self.event_dim == 0:
            return td.constraints.real
        return td.constraints.independent(td.constraints.real, self.event_dim)

    @td.constraints.dependent_property(is_discrete=False)
    def codomain(self):
        if self.event_dim == 0:
            return td.constraints.real
        return td.constraints.independent(td.constraints.real, self.event_dim)

    def _call(self, x: torch.Tensor) -> torch.Tensor:
        """
        Args:
            x: batch_shape + (input_dim)
        Returns:
            transformed_input batch_shape + (input_dim)
        """
        return self._piecewise_cdf(x, inverse=False)[0]

    def _inverse(self, y: torch.Tensor) -> torch.Tensor:
        """
        Args:
            y: batch_shape + (input_dim)
        Returns:
            transformed_input, batch_shape + (input_dim)
        """
        return self._piecewise_cdf(y, inverse=True)[0]

    def log_abs_det_jacobian(self, x: torch.Tensor, y: torch.Tensor) -> torch.Tensor:
        """
        Args:
            input: batch_shape + (input_dim)
        Returns:
            jacobian_log_determinant: batch_shape + (input_dim)
        """
        return self._piecewise_cdf(x, inverse=False)[1]



from typing import Any

import fsspec


def get_storage_options_for_path(path: str) -> dict[str, Any]:
    """Get storage options for a given path.

    This checks whether a path uses the abfs protocol and if so returns storage options for the Azure Blob Storage.

    Args:
        path: fsspec compatible path

    Returns:
        Dictionary of storage options for the given path.
    """
    fs, _ = fsspec.core.url_to_fs(path)
    if fs.protocol == "abfs":
        return {"az": {"anon": False}}
    return {}



from causica.distributions.adjacency import (
    AdjacencyDistribution,
    ConstrainedAdjacency,
    ConstrainedAdjacencyDistribution,
    ENCOAdjacencyDistribution,
    ENCOAdjacencyDistributionModule,
    ErdosRenyiDAGDistribution,
    ExpertGraphContainer,
    GibbsDAGPrior,
    ThreeWayAdjacencyDistribution,
)
from causica.distributions.distribution_module import DistributionModule
from causica.distributions.noise import (
    BernoulliNoise,
    BernoulliNoiseModule,
    CategoricalNoise,
    CategoricalNoiseModule,
    ContinuousNoiseDist,
    IndependentNoise,
    JointNoise,
    JointNoiseModule,
    Noise,
    NoiseModule,
    SplineNoise,
    SplineNoiseModule,
    UnivariateNormalNoise,
    UnivariateNormalNoiseModule,
    create_noise_modules,
    create_spline_dist_params,
)
from causica.distributions.sem_distribution import SEMDistribution, SEMDistributionModule



from typing import Optional

import pandas as pd
import torch
from tensordict import TensorDict


def convert_one_hot(
    data: TensorDict,
    one_hot_sizes: Optional[dict[str, int]] = None,
):
    """
    Args:
        data: A tensordict representing the underlying data
        one_hot_keys: A list of keys to convert to one hot encodings
    """
    # one hot encode the categorical variables
    if one_hot_sizes is None:
        return data

    new_data = data.clone()
    for key, num_classes in one_hot_sizes.items():
        cat_var = new_data[key]
        assert cat_var.shape == new_data.batch_size + (1,), "Only support 1D categorical values"
        new_data[key] = torch.nn.functional.one_hot(cat_var[..., 0].to(torch.long), num_classes=num_classes)
    return new_data


def tensordict_shapes(tds: TensorDict) -> dict[str, torch.Size]:
    """Return the shapes within the TensorDict without batch dimensions."""
    return {key: val.shape[len(tds.batch_size) :] for key, val in tds.items()}


def tensordict_from_pandas(df: pd.DataFrame) -> TensorDict:
    """
    Create a `TensorDict` from a pandas dataframe.

    It supports pandas `MultiIndex`s of depth 2 and uses the top level index
    to determine the number of keys in the dictionary.
    """
    if isinstance(df.columns, pd.MultiIndex):
        assert df.columns.nlevels == 2, "Only support MultiIndex of depth 2."
        data = {key: df[key].to_numpy() for key in df.columns.get_level_values(0).unique()}
    else:
        # in the single index case, add a dimension to the series, so they have shape [batch, 1]
        data = {key: df[key].to_numpy()[:, None] for key in df.columns}

    return TensorDict(data, batch_size=torch.Size([len(df)]))


def identity(x):
    """
    An identity function.

    This is needed to override the default collate function for the dataloaders,
    which limit the datatypes you can return and we have `TensorDict` etc.
    """
    return x



from collections import defaultdict

import pandas as pd
import torch
from tensordict import TensorDict
from torch.utils.data import DataLoader

from causica.datasets.causica_dataset_format import convert_variable_types_to_enum, tensordict_from_variables_metadata
from causica.datasets.standardizer import fit_standardizer
from causica.datasets.tensordict_utils import identity, tensordict_shapes
from causica.datasets.variable_types import VariableTypeEnum
from causica.lightning.data_modules.deci_data_module import DECIDataModule


class BasicDECIDataModule(DECIDataModule):
    """A datamodule interface for a dataframe and variable specification."""

    def __init__(
        self,
        dataframe: pd.DataFrame,
        variables_spec: dict,
        batch_size: int,
        normalize: bool = False,
        dataset_name: str = "anonymous_dataset",
    ):
        super().__init__()
        self.dataset_df = dataframe
        self.variables_metadata = convert_variable_types_to_enum(variables_spec)
        self._dataset_name = dataset_name
        self.normalize = normalize
        self.batch_size = batch_size

        self._dataset_train = tensordict_from_variables_metadata(
            self.dataset_df.to_numpy(), self.variables_metadata["variables"]
        )
        self._variable_shapes = tensordict_shapes(self._dataset_train)
        self._variable_types = {var["group_name"]: var["type"] for var in self.variables_metadata["variables"]}
        self._column_names = defaultdict(list)
        for variable in self.variables_metadata["variables"]:
            self._column_names[variable["group_name"]].append(variable["name"])

        if self.normalize:
            continuous_keys = [k for k, v in self._variable_types.items() if v == VariableTypeEnum.CONTINUOUS]
            self.normalizer = fit_standardizer(self._dataset_train.select(*continuous_keys))

            transform = self.normalizer()
            self._dataset_train = transform(self._dataset_train)

    @property
    def dataset_name(self) -> TensorDict:
        return self._dataset_name

    @property
    def dataset_train(self) -> TensorDict:
        return self._dataset_train

    @property
    def dataset_test(self) -> TensorDict:
        raise RuntimeError("Not defined")

    @property
    def variable_shapes(self) -> dict[str, torch.Size]:
        """Get the shape of each variable in the dataset."""
        return self._variable_shapes

    @property
    def variable_types(self) -> dict[str, VariableTypeEnum]:
        """Get the type of each variable in the dataset."""
        return self._variable_types

    @property
    def column_names(self) -> dict[str, list[str]]:
        """Get a map of the variable names and the corresponding columns of the original dataset."""
        return self._column_names

    def train_dataloader(self):
        return DataLoader(
            dataset=self.dataset_train,
            collate_fn=identity,
            batch_size=self.batch_size,
            shuffle=True,
            drop_last=True,
        )



from typing import Iterable, Optional, Union

import torch
from tensordict import TensorDict

from causica.datasets.causica_dataset_format import CounterfactualWithEffects
from causica.distributions.transforms import JointTransform
from causica.sem.distribution_parameters_sem import DistributionParametersSEM
from causica.sem.structural_equation_model import SEM, counterfactual


def eval_counterfactual_outcome_per_variable_rmse(
    sems: Iterable[SEM],
    counterfactual_data: CounterfactualWithEffects,
    grouped_variable_names: Optional[dict[str, list[str]]] = None,
    standardizer: Optional[JointTransform] = None,
) -> TensorDict:
    """Evaluate the counterfacual rmses of a model.

    Args:
        sems: An iterable of structural equation models to evaluate the ITE RMSE of
        counterfactual_data: Data of true counterfactuals to use for evaluation.
        grouped_variable_names: Optional dictionary that holds the names of the variables in each group. If given,
            the variables are evaluated individually. Otherwise, the variables are evaluated groupwise.
        standardizer: Standardizer that is used to invert the predictions and loaded data to convery the metrics into
            the space of the original data.

    Returns:
        Dict of RMSEs for each effect variable we're interested in
    """
    intervention, _, effects = counterfactual_data

    # generate samples from the intervened distribution and the base distribution
    stacked: TensorDict = torch.stack(
        [counterfactual(sem, intervention.factual_data, intervention.intervention_values) for sem in sems]
    )
    generated_cf_outcomes = stacked.apply(
        lambda v: v.mean(axis=0), batch_size=intervention.factual_data.batch_size, inplace=False
    )
    true_counterfactual_outcome = intervention.counterfactual_data

    if standardizer is not None:
        generated_cf_outcomes = standardizer.inv(generated_cf_outcomes)
        true_counterfactual_outcome = standardizer.inv(true_counterfactual_outcome)

    if grouped_variable_names is None:
        # calculate the rmse metrics, one for each group
        rmses = TensorDict(
            {key: rmse(generated_cf_outcomes[key], true_counterfactual_outcome[key]) for key in effects},
            batch_size=torch.Size(),
        )
    else:
        # calculate the rmse metrics, one for each column/single variable
        rmses = TensorDict(
            {
                col_name: rmse(generated_cf_outcomes[key][:, idx], true_counterfactual_outcome[key][:, idx])
                for key in effects
                for idx, col_name in enumerate(grouped_variable_names[key])
            },
            batch_size=torch.Size(),
        )

    return rmses


def eval_observational_per_variable_rmse_and_accuracy(
    sems: Union[list[DistributionParametersSEM], tuple[DistributionParametersSEM]],
    observations: TensorDict,
    continuous_variables: list[str],  # keys to variables to be interpreted as categorical
    binary_variables: list[str],  # keys to variables to be interpreted as binary
    categorical_variables: list[str],  # keys to variables to be interpreted as categorical
    grouped_variable_names: Optional[dict[str, list[str]]] = None,
    standardizer: Optional[JointTransform] = None,
) -> tuple[dict[str, torch.Tensor], dict[str, torch.Tensor]]:
    """Calculates the RMSE and accuracy of the predictions of a model.

    Args:
        sems: An iterable of structural equation models to evaluate the ITE RMSE of
        observations: Observational data to evaluate
        continuous_variables: Keys of the continuous variables
        binary_variables: Keys of the binary variables
        categorical_variables: Keys of the categorical variables
        grouped_variable_names: Optional dictionary that holds the names of the variables in each group. If given,
            the variables are evaluated individually. Otherwise, the variables are evaluated groupwise.
        standardizer: Standardizer that is used to invert the predictions and loaded data to convery the metrics into
            the space of the original data.

    Returns:
        Dict holding the accuracy metrics for each node
    """
    assert all(
        set(sem.node_names) == set(observations.keys()) for sem in sems
    ), f"observations must be compatible with SEMs: got {set(observations.keys())} but expected {set(list(sems)[0].node_names)}"

    assert set(continuous_variables + binary_variables + categorical_variables) == set(
        observations.keys()
    ), f"observations must be compatible with variables: got {set(observations.keys())} but expected {set(continuous_variables + binary_variables + categorical_variables)}"
    assert set(continuous_variables).isdisjoint(
        set(binary_variables)
    ), f"continuous and binary variables must be disjoint: got {set(continuous_variables).intersection(set(binary_variables))}"
    assert set(continuous_variables).isdisjoint(
        set(categorical_variables)
    ), f"continuous and categorical variables must be disjoint: got {set(continuous_variables).intersection(set(categorical_variables))}"
    assert set(binary_variables).isdisjoint(
        set(categorical_variables)
    ), f"binary and categorical variables must be disjoint: got {set(binary_variables).intersection(set(categorical_variables))}"

    stacked: TensorDict = torch.stack([sem.func(observations, sem.graph) for sem in sems])
    mean_predictions = stacked.apply(lambda v: v.mean(axis=0), batch_size=observations.batch_size, inplace=False)

    if standardizer is not None:
        mean_predictions = standardizer.inv(mean_predictions)
        observations = standardizer.inv(observations)

    rmses = {}
    accuraccies = {}
    if grouped_variable_names is None:
        # Calculate metrics for each group
        for var in continuous_variables:
            rmses[f"{var}"] = rmse(mean_predictions.get(var), observations.get(var))

        for var in binary_variables:
            accuraccies[f"{var}"] = binary_accuracy(mean_predictions.get(var), observations.get(var))

        for var in categorical_variables:
            accuraccies[f"{var}"] = categorical_accuracy(mean_predictions.get(var), observations.get(var))
    else:
        # Calculate metrics for each variable in each group
        for group_name in continuous_variables:
            for i, var in enumerate(grouped_variable_names[group_name]):
                rmses[f"{var}"] = rmse(
                    mean_predictions.get(group_name)[:, i][:, None], observations.get(group_name)[:, i][:, None]
                )

        for group_name in binary_variables:
            for i, var in enumerate(grouped_variable_names[group_name]):
                accuraccies[f"{var}"] = binary_accuracy(
                    mean_predictions.get(group_name)[:, i], observations.get(group_name)[:, i]
                )

        # Categorical variables are one-hot encoded, so we need to calculate the accuracy for each group as we do not
        # support grouped categorical variables of shape [batch_size, num_groups, num_dims_for_node]
        for group_name in categorical_variables:
            accuraccies[f"{group_name}"] = categorical_accuracy(
                mean_predictions.get(group_name), observations.get(group_name)
            )

    return rmses, accuraccies


def binary_accuracy(logits: torch.Tensor, target: torch.Tensor) -> torch.Tensor:
    """Calculate the accuracy of a prediction for a binary variable.

    Args:
        logits: Tensor of logits [batch_size] or [batch_size, num_dims_for_node]
        target: Tensor of targets [batch_size] or [batch_size, num_dims_for_node]

    Returns:
        Accuracy of the prediction
    """
    return torch.mean(((torch.sigmoid(logits) > 0.5) == target).float())


def categorical_accuracy(logits: torch.Tensor, target: torch.Tensor, target_is_onehot: bool = True) -> torch.Tensor:
    """Calculate the accuracy of a prediction for a categorical variable.

    Args:
        logits: Tensor of logits [batch_size, num_classes]
        target: Tensor of targets [batch_size, num_classes] (if target_is_onehot is True) or [batch_size] (otherwise)
        target_is_onehot: Whether the target is one hot. Defaults to True.

    Returns:
        Accuracy of the prediction
    """
    if target_is_onehot:
        target = torch.argmax(target, -1)

    prediction = torch.argmax(logits, dim=-1)

    return torch.mean((prediction == target).float())


def rmse(prediction: torch.Tensor, target: torch.Tensor) -> torch.Tensor:
    """Calculate the RMSE of a prediction. This will sum over all dimensions except the batch dimension.

    Args:
        prediction: Tensor of predictions [batch_size, num_dims_for_node]
        target: Tensor of targets [batch_size, num_dims_for_node]

    Returns:
        RMSE of the prediction
    """
    return torch.sqrt(torch.mean(torch.sum((prediction - target) ** 2, -1)))



from typing import Optional

import numpy as np
import torch
import torch.distributions as td

from causica.distributions.adjacency.adjacency_distributions import AdjacencyDistribution
from causica.triangular_transformations import fill_triangular


class ErdosRenyiDAGDistribution(AdjacencyDistribution):
    """
    An adjacency distribution for sampling Directed Acyclic Graphs using Erdos Renyi.
    """

    support = td.constraints.independent(td.constraints.boolean, 1)

    def __init__(self, num_nodes: int, probs: torch.Tensor, validate_args: Optional[bool] = None):
        """
        Args:
            num_nodes: the number of nodes in the DAGs to be sampled
            probs: A tensor of the probability that an edge exists between 2 nodes of shape batch_shape
        """
        assert num_nodes > 0, "Number of nodes in the graph must be greater than 0"
        super().__init__(num_nodes=num_nodes, validate_args=validate_args)
        self.probs = probs
        self.num_low_tri = num_nodes * (num_nodes - 1) // 2
        expanded_probs = probs[..., None].expand(
            *probs.shape + (self.num_low_tri,)
        )  # shape batch_shape + (num_low_tri)
        self.bern_dist = td.Independent(td.Bernoulli(probs=expanded_probs), reinterpreted_batch_ndims=1)
        self.np_rng = np.random.default_rng()

    def relaxed_sample(self, sample_shape: torch.Size = torch.Size(), temperature: float = 0.0) -> torch.Tensor:
        raise NotImplementedError

    def sample(self, sample_shape: torch.Size = torch.Size()) -> torch.Tensor:
        """
        Sample a binary adjacency matrix from the underlying distribution.

        Args:
            sample_shape: the shape of the samples to return
        Returns:
            A tensor of shape sample_shape + batch_shape + (num_nodes, num_nodes)
        """
        # generate the lower triangle shape [..., n, n]
        low_tri = fill_triangular(self.bern_dist.sample(sample_shape=sample_shape))

        # generate a random permutation matrix
        aranges = np.tile(np.arange(self.num_nodes), sample_shape + self.probs.shape + (1,))  # shape [..., n]
        np_perms = torch.tensor(self.np_rng.permuted(aranges, axis=-1))  # a batch of rearranged [0, 1, 2... n]
        # one hot the last dimension to create a tensor of shape [..., n, n]
        perms = torch.nn.functional.one_hot(np_perms, num_classes=self.num_nodes).to(dtype=low_tri.dtype)

        return torch.einsum("...ij,...jk,...lk->...il", perms, low_tri, perms)

    def entropy(self) -> torch.Tensor:
        raise NotImplementedError

    @property
    def mean(self) -> torch.Tensor:
        raise NotImplementedError

    @property
    def mode(self) -> torch.Tensor:
        """
        Return the mode of the underlying distribution.

        This will be an adjacency matrix.

        For this distribution, there are many modes since each permutation of nodes is equally likely.
        We return the mode corresponding to the "default" ordering.

        There are 2 possibilities:
            p > 0.5: A lower triangular matrix of ones
            p <= 0.5: A matrix of zeros

        Returns:
            A tensor of shape batch_shape + (num_nodes, num_nodes)
        """
        return fill_triangular(torch.nan_to_num(self.bern_dist.mode, nan=0.0))

    def log_prob(self, value: torch.Tensor) -> torch.Tensor:
        raise NotImplementedError



from typing import Optional

import torch
import torch.distributions as td
import torch.nn.functional as F
from torch import nn
from torch.distributions import OneHotCategorical

from causica.distributions.noise.noise import Noise, NoiseModule


class CategoricalNoise(OneHotCategorical, Noise):
    def __init__(self, delta_logits: torch.Tensor, base_logits: torch.Tensor):
        """
        A Categorical distribution with parameters defined by base_logits and self.delta_logits (ICGNN predictions).

        Args:
            delta_logits: Tensor with shape [sample_shape, event_shape]
            base_logits: Tensor with shape [event_shape] where event_shape shows the number of categories for this node
        """
        self.delta_logits = delta_logits
        super().__init__(logits=base_logits + delta_logits, validate_args=False)

    def sample_to_noise(self, samples: torch.Tensor) -> torch.Tensor:
        """
        Transform from the sample observations to corresponding noise variables.

        This will draw from the noise posterior given the observations

        A posterior sample of the Gumbel noise random variables given observation x and probabilities
        `self.base_logits + logit_deltas`.

        This methodology is described in https://arxiv.org/pdf/1905.05824.pdf.
        See https://cmaddis.github.io/gumbel-machinery for derivation of Gumbel posteriors.
        For a derivation of this exact algorithm using softplus, see https://www.overleaf.com/8628339373sxjmtvyxcqnx.

        Args:
            samples: Tensor of shape sample_shape + batch_shape + event_shape
        Returns:
            The generated samples with shape sample_shape + batch_shape + event_shape
        """

        device = self.delta_logits.device
        dist = td.Gumbel(torch.tensor(0.0, device=device), torch.tensor(1.0, device=device))
        top_sample = dist.sample(samples.shape[:-1] + (1,)) + self.logits.logsumexp(-1, keepdim=True)
        lower_samples = dist.sample(samples.shape) + self.logits
        lower_samples[samples == 1] = float("inf")
        return top_sample - F.softplus(top_sample - lower_samples) - self.delta_logits

    def noise_to_sample(self, noise: torch.Tensor) -> torch.Tensor:
        """
        Generate samples using the given exogenous noise.

        Args:
            noise: noise variable with shape sample_shape + batch_shape.
        Returns:
            The generated samples with shape sample_shape + batch_shape + event_shape
        """
        x = self.delta_logits + noise
        maxes = torch.max(x, dim=-1, keepdim=True)
        return (x >= maxes[0]).float()


class CategoricalNoiseModule(NoiseModule[CategoricalNoise]):
    """Represents a CategoricalNoise distribution with learnable logits."""

    def __init__(self, num_classes: int):
        """
        Args:
            num_classes: Number of classes.
        """
        super().__init__()
        self.base_logits = nn.Parameter(torch.zeros(num_classes))

    def forward(self, x: Optional[torch.Tensor] = None) -> CategoricalNoise:
        if x is None:
            x = torch.zeros_like(self.base_logits)
        return CategoricalNoise(delta_logits=x, base_logits=self.base_logits)



"""
Wrapper around torch.distributions.transforms to allow for joint transforms on TensorDicts.
"""

import torch
import torch.distributions as td
from tensordict import TensorDict


class JointTransform(td.Transform):
    """A joint transform that applies a different transform to each key in the TensorDict.

    This is heavily inspired by the `torch.distributions.transforms.StackTransform` class.
    See https://pytorch.org/docs/stable/distributions.html#torch.distributions.transforms.StackTransform
    """

    def __init__(self, transformations: dict[str, td.Transform], cache_size: int = 0):
        """
        Args:
            transformations: A dictionary of transforms, where the keys are the keys in the TensorDict
            cache_size: Size of cache. If zero, no caching is done. If one, the latest single value is cached.
                Only 0 and 1 are supported.
        """
        assert all(
            isinstance(t, td.Transform) for t in transformations.values()
        ), f"All transformations must be of type torch.distributions.Transform, but are {[type(t) for t in transformations.values()]}."
        if cache_size:
            transformations = {key: t.with_cache(cache_size) for key, t in transformations.items()}
        super().__init__(cache_size=cache_size)

        self.transformations = transformations

    def _call(self, x: TensorDict) -> TensorDict:
        if not set(self.transformations.keys()).issubset(x.keys()):
            raise ValueError("All keys in transformations must be in x.")

        return x.clone().update({key: transform(x[key]) for key, transform in self.transformations.items()})

    def _inverse(self, y: TensorDict) -> TensorDict:
        # We cannot use ._inv as pylint complains with E202: _inv is hidden because of `self._inv = None`
        # in td.Transform.__init__
        if not set(self.transformations.keys()).issubset(y.keys()):
            raise ValueError("All keys in transformations must be in y.")

        return y.clone().update({key: transform.inv(y[key]) for key, transform in self.transformations.items()})

    def log_abs_det_jacobian(self, x: TensorDict, y: TensorDict) -> torch.Tensor:
        if set(x.keys()) != set(y.keys()):
            raise ValueError("x and y must have the same keys.")

        if not set(self.transformations.keys()).issubset(x.keys()):
            raise ValueError("All keys in transformations must be in x and y.")

        return x.clone().update(
            {
                key: self.transformations[key].log_abs_det_jacobian(x[key], y[key])
                if key in self.transformations
                else torch.ones_like(x[key])
                for key in x.keys()
            }
        )

    @property
    def bijective(self):
        return all(t.bijective for t in self.transformations.values())

    @property
    def domain(self):
        return {key: t.domain for key, t in self.transformations.items()}

    @property
    def codomain(self):
        return {key: t.codomain for key, t in self.transformations.items()}



from functools import partial
from typing import Callable, Type

import torch
import torch.distributions as td

from causica.distributions.adjacency.adjacency_distributions import AdjacencyDistribution
from causica.distributions.distribution_module import DistributionModule


class ConstrainedAdjacencyDistribution(AdjacencyDistribution):
    """Adjacency distribution that applies hard constraints to a base distribution.

    Overrides elements produced from the base adjacency distribution with
    - 0 when the corresponding negative_constraint=0
    - 1 when the corresponding positive constraint=1
    - unmodified: all other elements
    """

    arg_constraints = {"positive_constraints": td.constraints.boolean, "negative_constraints": td.constraints.boolean}

    def __init__(
        self, dist: AdjacencyDistribution, positive_constraints: torch.Tensor, negative_constraints: torch.Tensor
    ):
        """
        Args:
            dist (AdjacencyDistribution): Base distribution, event shape matching the postive and negative constraints.
            positive_constraints (torch.Tensor): Positive constraints. 1 means edge is present.
            negative_constraints (torch.Tensor): Negative constraints. 0 means edge is not present.
        """
        if not dist.event_shape == positive_constraints.shape == negative_constraints.shape:
            raise ValueError("The constraints must match the event shape of the distribution.")
        self.dist = dist
        self.positive_constraints = positive_constraints
        self.negative_constraints = negative_constraints

        super().__init__(self.dist.num_nodes)

    def relaxed_sample(self, sample_shape: torch.Size = torch.Size(), temperature: float = 0.0) -> torch.Tensor:
        """
        Sample a binary adjacency matrix from the relaxed distribution and apply constraints.

        Args:
            sample_shape: the shape of the samples to return
            temperature: The temperature of the relaxed distribution
        Returns:
            A tensor of shape sample_shape + batch_shape + (num_nodes, num_nodes)
        """
        return self._apply_constraints(self.dist.relaxed_sample(sample_shape=sample_shape, temperature=temperature))

    def sample(self, sample_shape: torch.Size = torch.Size()) -> torch.Tensor:
        """
        Sample a binary adjacency matrix from the underlying distribution and apply constraints.

        Args:
            sample_shape: the shape of the samples to return
        Returns:
            A tensor of shape sample_shape + batch_shape + (num_nodes, num_nodes)
        """
        return self._apply_constraints(self.dist.sample(sample_shape=sample_shape))

    @property
    def mean(self) -> torch.Tensor:
        """
        Return the mean of the underlying distribution and applies the constraints.

        This will be a matrix with all entries in the interval [0, 1].

        Returns:
            A tensor of shape batch_shape + (num_nodes, num_nodes)
        """
        return self._apply_constraints(self.dist.mean)

    @property
    def mode(self) -> torch.Tensor:
        """
        Return the mode of the underlying distribution and applies the constraints.

        This will be an adjacency matrix.

        Returns:
            A tensor of shape batch_shape + (num_nodes, num_nodes)
        """
        return self._apply_constraints(self.dist.mode)

    def entropy(self) -> torch.Tensor:
        """
        Return the entropy of the underlying distribution.

        NOTE: This does not account for the constraints.

        Returns:
            A tensor of shape batch_shape, with the entropy of the distribution
        """
        return self.dist.entropy()

    def log_prob(self, value: torch.Tensor) -> torch.Tensor:
        """
        Get the log probability of each tensor from the sample space

        NOTE: This does not account for the constraints.

        Args:
            value: a binary matrix of shape value_shape + batch_shape + (n, n)
        Returns:
            A tensor of shape value_shape + batch_shape, with the log probabilities of each tensor in the batch.
        """
        # entropy might need to be modified to account for constraints
        return self.dist.log_prob(value)

    def _apply_constraints(self, G: torch.Tensor) -> torch.Tensor:
        """Return G with the positive and negative constraints applied."""
        return 1.0 - (1.0 - G * self.negative_constraints) * (~self.positive_constraints)


def get_graph_constraint(graph_constraint_matrix: torch.Tensor) -> tuple[torch.Tensor, torch.Tensor]:
    """Converts graph constraint matrix into a positive and negative matrix for easier usage.

    Args:
        graph_constraint_matrix: Graph constraints: 0 = no edge, 1 = edge, nan: no constraint. Should be a square matrix
                                 with the lengths matching the number of nodes in the graph.

    Returns:
        A tuple of (positive_constraints, negative_constraint). See ConstrainedAdjacencyDistribution for their
        interpretation.
    """
    assert graph_constraint_matrix.ndim == 2, "Constraint matrix must be 2D."
    assert graph_constraint_matrix.shape[0] == graph_constraint_matrix.shape[1], "Constraint matrix must be square."
    # Mask self-edges
    mask = ~torch.eye(graph_constraint_matrix.shape[0], dtype=torch.bool, device=graph_constraint_matrix.device)

    positive_constraints = mask * torch.nan_to_num(graph_constraint_matrix, nan=0).to(dtype=torch.bool)
    negative_constraints = torch.nan_to_num(graph_constraint_matrix, nan=1).to(dtype=torch.bool)
    return positive_constraints, negative_constraints


def _create_distribution(
    dist_class: Type[AdjacencyDistribution], *args, graph_constraint_matrix: torch.Tensor, **kwargs
) -> ConstrainedAdjacencyDistribution:
    """Utility function for generating a constrained adjacency distribution with a base distribution.

    Args:
        dist_class (Type[AdjacencyDistribution]): Type of the base adjacency distribution.
        graph_constraint_matrix: Graph constraints: 0 = no edge, 1 = edge, nan: no constraint. Must match the event
                                 shape of the distribution.

    Returns:
        ConstrainedAdjacencyDistribution: Constrained adjacency distribution.
    """

    positive_constraints, negative_constraints = get_graph_constraint(graph_constraint_matrix)

    dist = dist_class(*args, **kwargs)
    return ConstrainedAdjacencyDistribution(
        dist, positive_constraints=positive_constraints, negative_constraints=negative_constraints
    )


def constrained_adjacency(
    dist_class: Type[AdjacencyDistribution],
) -> Callable[..., ConstrainedAdjacencyDistribution]:
    """Utility function that returns a function constructing a constrained adjacency distribution.

    Args:
        dist_class: Type of the base adjacency distribution.
        graph_constraint_matrix: Graph constraints: 0 = no edge, 1 = edge, nan: no constraint. Must match the event
                                 shape of the distribution.

    Returns:
        Callable[..., ConstrainedAdjacencyDistribution]: Utility function creating a ConstrainedAdjacencyDistribution.
    """

    return partial(
        _create_distribution,
        dist_class=dist_class,
    )


class ConstrainedAdjacency(DistributionModule[AdjacencyDistribution]):
    """A constrained adjacency distribution module where certain parts edges of in the adjacency matrix are locked."""

    def __init__(
        self, adjacency_distribution: DistributionModule[AdjacencyDistribution], graph_constraint_matrix: torch.Tensor
    ):
        """
        Args:
            adjacency_distribution: Underlying adjacency distribution module.
            graph_constraint_matrix: Constraint matrix with edges defined according to `get_graph_constraint`.
        """
        super().__init__()
        self.adjacency_distribution = adjacency_distribution
        positive_constraints, negative_constraints = get_graph_constraint(graph_constraint_matrix)
        self.positive_constraints: torch.Tensor
        self.negative_constraints: torch.Tensor
        self.register_buffer("positive_constraints", positive_constraints)
        self.register_buffer("negative_constraints", negative_constraints)

    def forward(self) -> ConstrainedAdjacencyDistribution:
        return ConstrainedAdjacencyDistribution(
            self.adjacency_distribution(),
            positive_constraints=self.positive_constraints,
            negative_constraints=self.negative_constraints,
        )



from typing import Optional

import torch
import torch.distributions as td
import torch.nn.functional as F
from torch import nn
from torch.distributions.utils import logits_to_probs

from causica.distributions.adjacency.adjacency_distributions import AdjacencyDistribution
from causica.distributions.distribution_module import DistributionModule
from causica.distributions.gumbel_binary import gumbel_softmax_binary
from causica.triangular_transformations import fill_triangular


class ENCOAdjacencyDistribution(AdjacencyDistribution):
    """
    The parameterization from the ENCO paper (https://arxiv.org/pdf/2107.10483.pdf).
    For each edge, it parameterizes the existence and orientation separately.

    We represent the orientation logit probabilities (θ) as a vector of length n(n - 1) / 2, (the strictly lower triangle)
    We represent the existence logit probabilities (γ) as a matrix of size (n, n), where the main diagonal is ignored.

    The probability of an edge existing is pᵢⱼ = σ(γᵢⱼ)σ(θᵢⱼ), where:
        σ is the sigmoid function
        θᵢⱼ is the orientation logit and θᵢⱼ = -θⱼᵢ, so the probabilities sum to one.
        γᵢⱼ is the logit the edge exists and is independent for each i, j. Note γᵢⱼ need not equal -γⱼᵢ

    The methods for this class constrct logit(pᵢⱼ), which is then a matrix of independent Bernoulli.
    """

    arg_constraints = {"logits_exist": td.constraints.real, "logits_orient": td.constraints.real}

    def __init__(
        self,
        logits_exist: torch.Tensor,
        logits_orient: torch.Tensor,
        validate_args: Optional[bool] = None,
    ):
        """
        Args:
            logits_exist: The logits for the existence of an edge, a batch_shape + (n, n) tensor
            logits_orient: The logits for the orientation of each edge, a tensor with shape batch_shape + n(n-1)/2
            validate_args: Whether to validate the arguments. Passed to the superclass
        """
        num_nodes = logits_exist.shape[-1]

        if validate_args:
            assert len(logits_exist.shape) >= 2, "Logits_exist must be a matrix, batching is not supported"
            assert logits_exist.shape[-2:] == (num_nodes, num_nodes), "Invalid logits_exist shape"
            assert len(logits_orient.shape) >= 1, "Logits_exist must be 1 dimensional, batching is not supported"
            assert logits_orient.shape[-1] == (num_nodes * (num_nodes - 1)) // 2, "Invalid logits_orient shape"
            assert logits_exist.device == logits_orient.device, "Logits must exist on the same device"

        self.logits_exist = logits_exist
        self.logits_orient = logits_orient

        super().__init__(num_nodes, validate_args=validate_args)

    def _get_independent_bernoulli_logits(self) -> torch.Tensor:
        """
        Construct the matrix logit(pᵢⱼ).


        See the class docstring.
        We use the following derivation
            logit(pᵢⱼ) = - log(pᵢⱼ⁻¹ - 1)
                       = - log((1 + exp(-γᵢⱼ))(1 + exp(-θᵢⱼ)) - 1)
                       = - log(exp(-γᵢⱼ) + exp(-θᵢⱼ) + exp(-θᵢⱼ - γᵢⱼ))
                       = - logsumexp([-γᵢⱼ, -θᵢⱼ, -θᵢⱼ - γᵢⱼ])
        """
        # (..., num_nodes, num_nodes)
        neg_theta = fill_triangular(self.logits_orient, upper=True) - fill_triangular(self.logits_orient, upper=False)
        return -torch.logsumexp(
            torch.stack([-self.logits_exist, neg_theta, neg_theta - self.logits_exist], dim=-1), dim=-1
        )

    @staticmethod
    def base_dist(logits: torch.Tensor) -> td.Distribution:
        """A matrix of independent Bernoulli distributions."""
        return td.Independent(td.Bernoulli(logits=logits), 2)

    def relaxed_sample(self, sample_shape: torch.Size = torch.Size(), temperature: float = 0.0) -> torch.Tensor:
        """
        Sample from a relaxed distribution. We use a Gumbel Softmax.

        Args:
            sample_shape: the shape of the samples to return
        Returns:
            A tensor of shape sample_shape + batch_shape + (num_nodes, num_nodes)
        """
        logits = self._get_independent_bernoulli_logits()
        expanded_logits = logits.expand(*(sample_shape + logits.shape))
        samples = gumbel_softmax_binary(logits=expanded_logits, tau=temperature, hard=True)
        return samples * (1.0 - torch.eye(self.num_nodes, device=logits.device))

    def sample(self, sample_shape: torch.Size = torch.Size()) -> torch.Tensor:
        """
        Sample from the underyling independent Bernoulli distribution.

        Gradients will not flow through this method, use relaxed_sample instead.

        Args:
            sample_shape: the shape of the samples to return
        Returns:
            A tensor of shape sample_shape + batch_shape + (num_nodes, num_nodes)
        """
        logits = self._get_independent_bernoulli_logits()
        samples = self.base_dist(logits).sample(sample_shape=sample_shape)
        torch.diagonal(samples, dim1=-2, dim2=-1).zero_()  # zero the diagonal elements
        return samples

    def entropy(self) -> torch.Tensor:
        """
        Get the entropy of the independent Bernoulli distribution.

        Returns:
            A tensor of shape batch_shape, with the entropy of the distribution
        """
        logits = self._get_independent_bernoulli_logits()
        entropy = F.binary_cross_entropy_with_logits(logits, logits_to_probs(logits, is_binary=True), reduction="none")
        return torch.sum(entropy, dim=(-2, -1)) - torch.sum(torch.diagonal(entropy, dim1=-2, dim2=-1), dim=-1)

    @property
    def mean(self) -> torch.Tensor:
        """
        Return the mean of the underlying independent Bernoulli distribution.

        This will be a matrix with all entries in the interval [0, 1].

        Returns:
            A tensor of shape batch_shape + (num_nodes, num_nodes)
        """
        logits = self._get_independent_bernoulli_logits()
        return self.base_dist(logits).mean * (1.0 - torch.eye(self.num_nodes, device=logits.device))

    @property
    def mode(self) -> torch.Tensor:
        """
        Return the mode of the underlying independent Bernoulli distribution.

        This will be an adjacency matrix.

        Returns:
            A tensor of shape batch_shape + (num_nodes, num_nodes)
        """
        logits = self._get_independent_bernoulli_logits()
        # bernoulli mode can be nan for very small logits, favour sparseness and set to 0
        return torch.nan_to_num(self.base_dist(logits).mode, nan=0.0) * (
            1.0 - torch.eye(self.num_nodes, device=logits.device)
        )

    def log_prob(self, value: torch.Tensor) -> torch.Tensor:
        """
        Get the log probability of each tensor from the sample space

        Args:
            value: a binary matrix of shape value_shape + batch_shape + (n, n)
        Returns:
            A tensor of shape value_shape + batch_shape, with the log probabilities of each tensor in the batch.
        """
        # need to subtract the diagonal log probabilities
        logits = self._get_independent_bernoulli_logits()
        full_log_prob = self.base_dist(logits).log_prob(value)
        diag_log_prob = td.Independent(td.Bernoulli(logits=torch.diagonal(logits, dim1=-2, dim2=-1)), 1).log_prob(
            torch.diagonal(value, dim1=-2, dim2=-1)
        )
        return full_log_prob - diag_log_prob


class ENCOAdjacencyDistributionModule(DistributionModule[ENCOAdjacencyDistribution]):
    """Represents an `ENCOAdjacencyDistributionModule` distribution with learnable logits."""

    def __init__(self, num_nodes: int) -> None:
        super().__init__()
        self.logits_exist = nn.Parameter(torch.zeros(num_nodes, num_nodes), requires_grad=True)
        self.logits_orient = nn.Parameter(torch.zeros(int(num_nodes * (num_nodes - 1) / 2)), requires_grad=True)

    def forward(self) -> ENCOAdjacencyDistribution:
        return ENCOAdjacencyDistribution(logits_exist=self.logits_exist, logits_orient=self.logits_orient)



import fsspec
import numpy as np
import torch

from causica.distributions.adjacency import ExpertGraphContainer
from causica.fsspec_helpers import get_storage_options_for_path


class LoadedExpertGraphContainer(ExpertGraphContainer):
    """Loads an expert graph container from a file path."""

    def __init__(self, graph_path: str, confidence: float, scale: float):
        """
        Args:
            graph_path: fsspec compatible path to a saved graph, with np.nan indicating undecided values.
                This can be saved as a .npy file or a .csv file.
            confidence: See ExpertGraphContainer
            scale: See ExpertGraphContainer
        """
        storage_options = get_storage_options_for_path(graph_path)
        with fsspec.open(graph_path, **storage_options) as f:
            if graph_path.endswith(".csv"):
                prior_adj_matrix = np.loadtxt(f, delimiter=",")
            else:
                prior_adj_matrix = np.load(f)
        prior_mask = ~np.isnan(prior_adj_matrix)
        prior_adj_matrix = np.nan_to_num(prior_adj_matrix)
        super().__init__(
            dag=torch.tensor(prior_adj_matrix), mask=torch.tensor(prior_mask), confidence=confidence, scale=scale
        )



import abc
from typing import Generic, Optional, TypeVar

import torch
import torch.distributions as td

from causica.distributions.distribution_module import DistributionModule

SampleType = TypeVar("SampleType")


class Noise(Generic[SampleType], abc.ABC, td.Distribution):
    """
    Extend Distributions to allow the noise (usually unparametrized) to be extracted from samples and vice versa.

    This class is used for Counterfactuals, where we want to keep the noise fixed but fix some values of the SEM.
    """

    @abc.abstractmethod
    def sample_to_noise(self, samples: SampleType) -> SampleType:
        """
        Transform from the sample observations to corresponding noise variables.

        Args:
            samples: Tensor of shape sample_shape + batch_shape + event_shape
        Returns:
            The generated samples with shape sample_shape + batch_shape + event_shape
        """

    @abc.abstractmethod
    def noise_to_sample(self, noise: SampleType) -> SampleType:
        """
        Generate samples using the given exogenous noise.

        Args:
            noise: noise variable with shape sample_shape + batch_shape.
        Returns:
            The generated samples with shape sample_shape + batch_shape + event_shape
        """


BaseNoiseType_co = TypeVar("BaseNoiseType_co", bound=Noise, covariant=True)


class IndependentNoise(Generic[BaseNoiseType_co], td.Independent, Noise[torch.Tensor]):
    """Like `td.Idenpendent` but also forwards `Noise` specific methods."""

    base_dist: BaseNoiseType_co

    def __init__(
        self,
        base_distribution: BaseNoiseType_co,
        reinterpreted_batch_ndims: int,
        validate_args: Optional[bool] = None,
    ):
        super().__init__(
            base_distribution=base_distribution,
            reinterpreted_batch_ndims=reinterpreted_batch_ndims,
            validate_args=validate_args,
        )

    def sample_to_noise(self, samples: torch.Tensor) -> torch.Tensor:
        """
        Transform from the sample observations to corresponding noise variables.

        This just passes through to the underlying distribution.

        Args:
            samples: Tensor of shape sample_shape + batch_shape + event_shape
        Returns:
            The generated samples with shape sample_shape + batch_shape + event_shape
        """
        return self.base_dist.sample_to_noise(samples)

    def noise_to_sample(self, noise: torch.Tensor) -> torch.Tensor:
        """
        Generate samples using the given exogenous noise.

        This just passes through to the underlying distribution.

        Args:
            noise: noise variable with shape sample_shape + batch_shape.
        Returns:
            The generated samples with shape sample_shape + batch_shape + event_shape
        """
        return self.base_dist.noise_to_sample(noise)


NoiseType_co = TypeVar("NoiseType_co", bound=Noise, covariant=True)


class NoiseModule(abc.ABC, Generic[NoiseType_co], DistributionModule[NoiseType_co]):
    """Module for producing `Noise`, where sampling from the distribution corresponds to adding noise to the input.

    All subclasses must allow sampling without input, which should correspond as closely as possible to adding the noise
    to a compatible zero vector.
    """

    @abc.abstractmethod
    def forward(self, x=None) -> NoiseType_co:
        pass



from typing import Optional

import torch
import torch.distributions as td
import torch.nn.functional as F
from torch import nn

from causica.distributions.noise.noise import IndependentNoise, Noise, NoiseModule


class BernoulliNoise(td.Bernoulli, Noise):
    def __init__(self, delta_logits: torch.Tensor, base_logits: torch.Tensor):
        """
        A Bernoulli distribution with parameters defined by base_logits and x_hat (predictions for noiseless value).

        Args:
            delta_logits: Tensor with shape sample_shape + batch_shape. These are the predicted values.
            base_logits: Tensor with shape batch_shape
        """
        self.delta_logits = delta_logits
        super().__init__(logits=base_logits + delta_logits, validate_args=False)

    def sample_to_noise(self, samples: torch.Tensor) -> torch.Tensor:
        """
        Transform from the sample observations to corresponding noise variables.

        This will draw from the noise posterior given the observations

        A posterior sample of the Gumbel noise random variables given observation x and probabilities
        `self.base_logits + logit_deltas`.

        This methodology is described in https://arxiv.org/pdf/1905.05824.pdf.
        See https://cmaddis.github.io/gumbel-machinery for derivation of Gumbel posteriors.
        For a derivation of this exact algorithm using softplus, see https://www.overleaf.com/8628339373sxjmtvyxcqnx.

        Args:
            samples: Tensor of shape sample_shape + batch_shape + event_shape
        Returns:
            The generated samples with shape sample_shape + batch_shape + event_shape
        """
        assert (
            samples.shape == self.delta_logits.shape
        ), "The shape of the input does not match the shape of the logit_deltas"
        device = self.delta_logits.device
        dist = td.Gumbel(torch.tensor(0.0, device=device), torch.tensor(1.0, device=device))
        diff_sample = dist.sample(samples.shape) - dist.sample(samples.shape)  # sample_shape + batch_shape
        neg_log_prob_non_sampled = F.softplus(self.logits * samples - self.logits * (1 - samples))
        positive_sample = F.softplus(diff_sample + neg_log_prob_non_sampled)
        return positive_sample * samples - positive_sample * (1 - samples) - self.delta_logits

    def noise_to_sample(self, noise: torch.Tensor) -> torch.Tensor:
        """
        Generate samples using the given exogenous noise.

        Args:
            noise: noise variable with shape sample_shape + batch_shape.
        Returns:
            The generated samples with shape sample_shape + batch_shape + event_shape
        """
        return ((self.delta_logits + noise) > 0).float()

    @property
    def mode(self):
        """
        Override the default `mode` method to prevent it returning nan's.

        We favour sparseness, so if logit == 0, set the mode to be zero.
        """
        return (self.logits > 0).to(self.logits)


class BernoulliNoiseModule(NoiseModule[IndependentNoise[BernoulliNoise]]):
    """Represents a BernoulliNoise distribution with learnable logits."""

    def __init__(self, dim: int):
        """
        Args:
            dim: Number of dimensions (independent Bernouilli's).
        """
        super().__init__()
        self.base_logits = nn.Parameter(torch.zeros(dim))

    def forward(self, x: Optional[torch.Tensor] = None) -> IndependentNoise[BernoulliNoise]:
        if x is None:
            x = torch.zeros_like(self.base_logits)
        return IndependentNoise(BernoulliNoise(delta_logits=x, base_logits=self.base_logits), 1)



from collections import deque
from dataclasses import dataclass, field
from typing import Union

import numpy as np
import torch
from dataclasses_json import dataclass_json
from torch.optim import Optimizer


class AugLagLossCalculator:
    def __init__(self, init_alpha: float, init_rho: float):
        self.alpha = init_alpha
        self.rho = init_rho

    def __call__(self, objective: torch.Tensor, constraint: torch.Tensor) -> torch.Tensor:
        return objective + self.alpha * constraint + self.rho * constraint * constraint / 2


@dataclass_json
@dataclass(frozen=True)
class AugLagLRConfig:
    """
    Configuration parameters for the AuglagLR scheduler.

    lr_update_lag: Number of iterations to wait before updating the learning rate.
    lr_update_lag_best: Number of iterations to wait after the best model before updating the learning rate.
    lr_init_dict: Dictionary of intitialization parameters for every new inner optimization step.
        This must contain all parameter_groups for all optimizers
    aggregation_period: Aggregation period to compare the mean of the loss terms across this period.
    lr_factor: Learning rate update schedule factor (exponential decay).
    penalty_progress_rate: Number of iterations to wait before updating rho based on the dag penalty.
    safety_rho: Maximum rho that could be updated to.
    safety_alpha: Maximum alpha that could be udated to.
    max_lr_down: Maximum number of lr update times to decide inner loop termination.
    inner_early_stopping_patience: Maximum number of iterations to run after the best inner loss to terminate inner loop.
    max_outer_steps: Maximum number of outer update steps.
    patience_penalty_reached: Maximum number of outer iterations to run after the dag penalty has reached a good value.
    patience_max_rho: Maximum number of iterations to run once rho threshold is reached.
    penalty_tolerance: Tolerance of the dag penalty
    max_inner_steps: Maximum number of inner loop steps to run.

    """

    lr_update_lag: int = 500
    lr_update_lag_best: int = 250
    lr_init_dict: dict[str, float] = field(
        default_factory=lambda: {"vardist": 0.1, "icgnn": 0.0003, "noise_dist": 0.003}
    )
    aggregation_period: int = 20
    lr_factor: float = 0.1
    penalty_progress_rate: float = 0.65
    safety_rho: float = 1e13
    safety_alpha: float = 1e13
    max_lr_down: int = 3
    inner_early_stopping_patience: int = 500
    max_outer_steps: int = 100
    patience_penalty_reached: int = 5
    patience_max_rho: int = 3
    penalty_tolerance: float = 1e-5
    max_inner_steps: int = 3000


class AugLagLR:
    def __init__(self, config: AugLagLRConfig) -> None:
        """A Pytorch like scheduler which performs the Augmented Lagrangian optimization procedure, which consists of
        an inner loop which optimizes the objective for a fixed set of lagrangian parameters. The lagrangian parameters are
        annealed in the outer loop, according to a schedule as specified by the hyperparameters.

        Args:
            config: An `AugLagLRConfig` object containing the configuration parameters.
        """
        self.config = config

        self.outer_opt_counter = 0
        self.outer_below_penalty_tol = 0
        self.outer_max_rho = 0
        self._prev_lagrangian_penalty = np.inf
        self._cur_lagrangian_penalty = np.inf

        self.best_loss = np.inf
        self.last_lr_update_step = 0
        self.num_lr_updates = 0
        self.last_best_step = 0
        self.loss_tracker: deque = deque([], maxlen=config.aggregation_period)
        self.step_counter = 0
        self.epoch_counter = 0

    def _init_new_inner_optimisation(self):
        """Init the hyperparameters for a new inner loop optimization."""
        self.best_loss = np.inf
        self.last_lr_update_step = 0
        self.num_lr_updates = 0
        self.last_best_step = 0
        self.loss_tracker.clear()
        self.step_counter = 0
        self.epoch_counter = 0

    def _is_inner_converged(self) -> bool:
        """Check if the inner optimization loop has converged, based on maximum number of inner steps, number of lr updates.

        Returns:
            bool: Return True if converged, else False.
        """
        return (
            self.step_counter >= self.config.max_inner_steps
            or self.num_lr_updates >= self.config.max_lr_down
            or self.last_best_step + self.config.inner_early_stopping_patience <= self.step_counter
        )

    def _is_outer_converged(self) -> bool:
        """Check if the outer loop has converged based on the main outer opt counter, if the dag penalty is below a threshold
        or if the rho parameter has reached a limit"

        Returns:
            bool: Return True if converged, else False
        """
        return (
            self.outer_opt_counter >= self.config.max_outer_steps
            or self.outer_below_penalty_tol >= self.config.patience_penalty_reached
            or self.outer_max_rho >= self.config.patience_max_rho
        )

    def _enough_steps_since_last_lr_update(self) -> bool:
        """Check if enough steps have been taken since the previous learning rate update, based on the previous
        update step iteration.

        Returns:
            bool: indicating whether sufficient steps have occurred since the last update
        """
        return self.last_lr_update_step + self.config.lr_update_lag <= self.step_counter

    def _enough_steps_since_best_model(self) -> bool:
        """Check the number of iteration steps which have been passed after seeing the current best model.

        Returns:
            bool: Returns True if last iteration at which learning rate was
            updated and last best loss iteration is less than total steps, else False.
        """
        return self.last_best_step + self.config.lr_update_lag_best <= self.step_counter

    def _update_lr(self, optimizer: Union[Optimizer, list[Optimizer]]):
        """Update the learning rate of the optimizer(s) based on the lr multiplicative factor.

        Args:
            optimizer: Optimizers of auglag to be updated.
        """
        self.last_lr_update_step = self.step_counter
        self.num_lr_updates += 1

        if isinstance(optimizer, list):
            for opt in optimizer:
                for param_group in opt.param_groups:
                    param_group["lr"] *= self.config.lr_factor
        else:
            for param_group in optimizer.param_groups:
                param_group["lr"] *= self.config.lr_factor

    def _reset_lr(self, optimizer: Union[Optimizer, list[Optimizer]]):
        """Reset the learning rate of individual param groups from lr init dictionary.

        Args:
            optimizer: Optimizer(s) corresponding to all param groups.
        """
        self.last_lr_update_step = self.step_counter

        if isinstance(optimizer, list):
            for opt in optimizer:
                for param_group in opt.param_groups:
                    param_group["lr"] = self.config.lr_init_dict[param_group["name"]]

        else:
            for param_group in optimizer.param_groups:
                param_group["lr"] = self.config.lr_init_dict[param_group["name"]]

    def _update_lagrangian_params(self, loss: AugLagLossCalculator):
        """Update the lagrangian parameters (of the auglag routine) based on the dag constraint values observed.

        Args:
            loss: loss with lagrangian attributes rho and alpha to be updated.
        """
        if self._cur_dag_penalty < self.config.penalty_tolerance:
            self.outer_below_penalty_tol += 1
        else:
            self.outer_below_penalty_tol = 0

        if loss.rho > self.config.safety_rho:
            self.outer_max_rho += 1

        if self._cur_dag_penalty > self._prev_lagrangian_penalty * self.config.penalty_progress_rate:
            print(f"Updating rho, dag penalty prev: {self._prev_lagrangian_penalty: .10f}")
            loss.rho *= 10.0
        else:
            self._prev_lagrangian_penalty = self._cur_dag_penalty
            loss.alpha += loss.rho * self._cur_dag_penalty
            if self._cur_dag_penalty == 0.0:
                loss.alpha *= 5
            print(f"Updating alpha to: {loss.alpha}")
        if loss.rho >= self.config.safety_rho:
            loss.alpha *= 5

        loss.rho = min([loss.rho, self.config.safety_rho])
        loss.alpha = min([loss.alpha, self.config.safety_alpha])

    def _is_auglag_converged(self, optimizer: Union[Optimizer, list[Optimizer]], loss: AugLagLossCalculator) -> bool:
        """Checks if the inner and outer loops have converged. If inner loop is converged,
        it initilaizes the optimisation parameters for a new inner loop. If both are converged, it returns True.

        Args:
            optimizer: Optimizer(s) corresponding to different parameter groups on which auglag is being performed.
            loss: Auglag loss.

        Returns:
            bool: Returns True if both inner and outer have converged, else False
        """
        if self._is_inner_converged():
            if self._is_outer_converged():
                return True
            else:
                self._update_lagrangian_params(loss)
                self.outer_opt_counter += 1
                self._init_new_inner_optimisation()
                self._reset_lr(optimizer)
        elif self._enough_steps_since_last_lr_update() and self._enough_steps_since_best_model():
            self._update_lr(optimizer)

        return False

    def _check_best_loss(self):
        """Update the best loss based on the average loss over an aggregation period."""
        if len(self.loss_tracker) == self.config.aggregation_period:
            avg_loss = np.mean(self.loss_tracker)
            if avg_loss < self.best_loss:
                self.best_loss = avg_loss
                self.last_best_step = self.step_counter

    def step(
        self,
        optimizer: Union[Optimizer, list[Optimizer]],
        loss: AugLagLossCalculator,
        loss_value: float,
        lagrangian_penalty: float,
    ) -> bool:
        """The main update method to take one auglag inner step.

        Args:
            optimizer: Optimizer(s) corresponding to different param groups.
            loss: auglag loss with lagrangian parameters
            loss_value: the actual value of the elbo for the current update step.
            lagrangian_penalty: Dag penalty for the current update step.

        Returns:
            bool: if the auglag has converged (False) or not (True)
        """
        assert lagrangian_penalty >= 0, "auglag penalty must be non-negative"
        self.loss_tracker.append(loss_value)
        self._cur_dag_penalty = lagrangian_penalty
        self.step_counter += 1
        self._check_best_loss()
        return self._is_auglag_converged(optimizer=optimizer, loss=loss)



from typing import Optional, Union

import torch
import torch.distributions as td
from torch import nn

from causica.distributions.noise.noise import IndependentNoise, Noise, NoiseModule
from causica.distributions.noise.spline.rational_quadratic_transform import PiecewiseRationalQuadraticTransform

# Ordered inputs to `_create_composite_layer`
SplineParams = tuple[torch.Tensor, ...]


class SplineNoise(td.TransformedDistribution, Noise):
    """A Spline Based Noise Distribution.

    Parametrized as in in [Neural Spline Flows](https://arxiv.org/pdf/1906.04032.pdf).
    """

    def __init__(
        self,
        base_loc: torch.Tensor,
        base_scale: torch.Tensor,
        spline_transforms: list[Union[td.AffineTransform, td.ComposeTransform]],
    ):
        """
        Args:
            base_loc: Loc of base normal distribution.
            base_scale: Scale of base normal distribution.
            spline_transforms: Spline transforms, where the last transform bijects from noise to samples.
        """
        self.base_loc = base_loc
        self.base_scale = base_scale
        last_transform = spline_transforms[-1]
        if spline_transforms and not (isinstance(last_transform, td.AffineTransform) or not last_transform.parts):
            # `td.identity` is just a `td.ComposeTransform` with no `parts`. It's used when there's no output shift.
            raise TypeError(
                "The last transformation must be `td.AffineTransform` or td.identity, but was "
                f"`{type(last_transform)}`."
            )
        super().__init__(
            base_distribution=td.Normal(loc=self.base_loc, scale=self.base_scale),
            transforms=spline_transforms,
        )

    def sample_to_noise(self, samples: torch.Tensor) -> torch.Tensor:
        """
        Transform from the sample observations to corresponding noise variables.

        Args:
            samples: Tensor of shape sample_shape + batch_shape + event_shape
        Returns:
            The generated samples with shape sample_shape + batch_shape + event_shape
        """
        output_affine_transform = self.transforms[-1]
        return output_affine_transform.inv(samples)

    def noise_to_sample(self, noise: torch.Tensor) -> torch.Tensor:
        """
        Generate samples using the given exogenous noise.

        Args:
            noise: noise variable with shape sample_shape + batch_shape.
        Returns:
            The generated samples with shape sample_shape + batch_shape + event_shape
        """
        output_affine_transform = self.transforms[-1]
        return output_affine_transform(noise)


def create_spline_dist_params(
    dim: int,
    num_bins: int,
    flow_steps: int,
    knot_locations_scale: float,
    derivatives_scale: float,
) -> list[SplineParams]:
    """Create initial values for a spline distribution.

    Args:
        dim: Number of dimensions of the represented variable.
        num_bins: Number of spline bins.
        flow_steps: Number of flow steps.
        knot_locations_scale: Scale of random values used for `knot_locations` of `PiecewiseRationalQuadraticTransform`.
        derivatives_scale: Scale of random values for `derivatives` of `PiecewiseRationalQuadraticTransform`.

    Returns:
        A list of parameters for `CompositeSplineLayer`s.
    """
    param_list: list[SplineParams] = []
    for i in range(flow_steps + 1):
        log_scale = torch.zeros(dim)  # this will be exponentiated when passed to the spline distribution
        loc = torch.zeros(dim)
        if i == flow_steps:
            param_list.append((loc, log_scale))
        else:
            knot_locations = knot_locations_scale * torch.randn(dim, num_bins, 2)
            derivatives = derivatives_scale * torch.randn(dim, num_bins - 1)
            param_list.append((loc, log_scale, knot_locations, derivatives))
    return param_list


class CompositeSplineLayer(nn.Module):
    """A layer constructing an affine transformation potentially composed with a `PiecewiseRationalQuadraticTransform`.

    Encapsulates one of the layers or `flow_steps` of the [Neural Spline Flows](https://arxiv.org/pdf/1906.04032.pdf).
    """

    def __init__(
        self,
        init_loc: torch.Tensor,
        init_log_scale: torch.Tensor,
        init_knot_locations: Optional[torch.Tensor] = None,
        init_derivatives: Optional[torch.Tensor] = None,
    ):
        """
        Args:
            init_loc: Initial value of base normal distribution loc.
            init_scale: Initial value of base normal distribution scale.
            init_knot_locations: Initialization value for the knot locations of `PiecewiseRationalQuadraticTransform`.
            init_derivatives: Initialization value for the derivatives of `PiecewiseRationalQuadraticTransform`.
        """
        super().__init__()
        self.loc = nn.Parameter(init_loc)
        self.log_scale = nn.Parameter(init_log_scale)
        if (init_knot_locations is None) != (init_derivatives is None):
            raise ValueError("Either both or none of `knot_location` and `derivatives` must be set.")
        self.knot_locations = nn.Parameter(init_knot_locations) if init_knot_locations is not None else None
        self.derivatives = nn.Parameter(init_derivatives) if init_derivatives is not None else None

    def __call__(self, *args, **kwargs) -> td.Transform:
        return super().__call__(*args, **kwargs)

    def forward(self) -> td.Transform:
        affine = td.AffineTransform(loc=self.loc, scale=torch.exp(self.log_scale))
        if self.knot_locations is None or self.derivatives is None:  # Test both to ensure correct typing below
            return affine
        return td.ComposeTransform([affine, PiecewiseRationalQuadraticTransform(self.knot_locations, self.derivatives)])


class SplineNoiseModule(NoiseModule[IndependentNoise[SplineNoise]]):
    """Implements Neural Spline Flow noise with learnable parameters.

    See [Neural Spline Flows](https://arxiv.org/pdf/1906.04032.pdf).
    """

    def __init__(
        self,
        dim: int,
        num_bins: int = 8,
        flow_steps: int = 1,
        init_knot_locations_scale: float = 1e-2,
        init_derivatives_scale: float = 1e-2,
    ):
        """
        Args:
            dim: Number of dimensions of the represented variable.
            num_bins: Number of spline bins.
            flow_steps: Number of flow steps.
            init_knot_locations_scale: Scale of random initialization values for `knot_locations` of
                                       `PiecewiseRationalQuadraticTransform`.
            init_derivatives_scale: Scale of random initialization values for `derivatives` of
                                    `PiecewiseRationalQuadraticTransform`.
        """
        super().__init__()
        spline_params = create_spline_dist_params(
            dim=dim,
            num_bins=num_bins,
            flow_steps=flow_steps,
            knot_locations_scale=init_knot_locations_scale,
            derivatives_scale=init_derivatives_scale,
        )
        self.composite_spline_layers = nn.ModuleList(CompositeSplineLayer(*params) for params in spline_params)
        self.base_scale: torch.Tensor
        self.base_loc: torch.Tensor
        self.register_buffer("base_loc", torch.zeros(dim))
        self.register_buffer("base_scale", torch.ones(dim))

    def forward(self, x: Optional[torch.Tensor] = None) -> IndependentNoise[SplineNoise]:
        transforms = [spline_layer() for spline_layer in self.composite_spline_layers]
        if x is not None:
            transforms.append(td.AffineTransform(loc=x, scale=torch.ones_like(x, device=x.device)))
        else:
            # Maintain the old behavior of having this transform biject between noise and samples.
            transforms.append(td.identity_transform)

        spline_dist = SplineNoise(base_loc=self.base_loc, base_scale=self.base_scale, spline_transforms=transforms)
        return IndependentNoise(spline_dist, 1)



import torch
from tensordict import TensorDict

from causica.functional_relationships.functional_relationships import (
    FunctionalRelationships,
    sample_dict_to_tensor,
    tensor_to_sample_dict,
)


class LinearFunctionalRelationships(FunctionalRelationships):
    """
    A simple linear functional relationship.
    """

    def __init__(
        self,
        variables: dict[str, torch.Size],
        initial_linear_coefficient_matrix: torch.Tensor,
        trainable: bool = False,
    ) -> None:
        """
        Args:
            variables: Dict of node shapes (how many dimensions a variable has)
                Order corresponds to the order in graph(s).
            initial_linear_coefficient_matrix: the linear coefficients [output_shape, output_shape]
            trainable: whether the coefficient matrix should be learnable
        """
        super().__init__(variables)

        self.stacked_variable_masks = torch.nn.Parameter(
            torch.stack(list(self.variable_masks.values())).float(), requires_grad=False
        )

        assert initial_linear_coefficient_matrix.shape == (self.output_shape, self.output_shape)
        self.linear_coefficients = torch.nn.Parameter(initial_linear_coefficient_matrix, requires_grad=trainable)

    def forward(self, samples: TensorDict, graphs: torch.Tensor) -> TensorDict:
        """
        Args:
            samples: Batched inputs, size batch_size_x + (processed_dim_all).
            graphs: Weighted adjacency matrix, size batch_size_g + (n, n)
        Returns:
            A Dict of tensors of shape batch_shape_x + batch_shape_g + (processed_dim_all)
        """
        return tensor_to_sample_dict(
            self.linear_map(sample_dict_to_tensor(samples, self.variable_masks), graphs), self.variable_masks
        )

    def linear_map(self, samples: torch.Tensor, graph: torch.Tensor) -> torch.Tensor:
        """
        Applies the linear function to a concatenated tensor of samples.

        Args:
            samples: tensor of shape batch_shape_x + [n_cols]
            graph: tensor of shape batch_shape_g + [n_nodes, n_nodes]
        Returns:
            tensor of shape batch_shape_x + batch_shape_g + [n_cols]
        """
        batch_shape_x = samples.shape[:-1]
        batch_shape_g = graph.shape[:-2]

        masked_graph = torch.einsum(
            "ji,...jk,kl->...il", self.stacked_variable_masks, graph, self.stacked_variable_masks
        )

        graph_broad = masked_graph.expand(*(batch_shape_x + tuple([-1] * len(graph.shape))))
        target_shape = batch_shape_x + batch_shape_g + samples.shape[-1:]
        view_shape = batch_shape_x + (1,) * len(batch_shape_g) + samples.shape[-1:]
        # Shape batch_shape_x + batch_shape_g + (num_nodes, out_dim_g)
        samples_broad = samples.view(view_shape).expand(target_shape)

        return torch.einsum("...i,...ij->...j", samples_broad, graph_broad * self.linear_coefficients)



from typing import Deque

import numpy as np


class AverageMetricTracker:
    """A class to keep the smallest value of the rolling average of a metric over time.

    Args:
        averaging_period: Number of steps to average over.
    """

    def __init__(self, averaging_period: int = 10):
        self._averaging_period = averaging_period
        self.min_value = np.inf
        self.queue: Deque = Deque([], maxlen=self._averaging_period)
        self.rolling_sum = 0.0

    @property
    def average(self):
        return self.rolling_sum / len(self.queue)

    def step(self, value: float) -> bool:
        """Add a new value to the tracker."""
        removed_value = self.queue.popleft() if len(self.queue) == self._averaging_period else 0.0
        self.queue.append(value)
        self.rolling_sum += value - removed_value

        if (current_average := self.average) < self.min_value:
            self.min_value = current_average
            return True
        return False

    def reset(self):
        """Reset the tracker."""
        self.min_value = np.inf
        self.queue = Deque([], maxlen=self._averaging_period)
        self.rolling_sum = 0.0



import torch
import torch.distributions as td
from torch.distributions.constraints import Constraint

from causica.distributions.adjacency import AdjacencyDistribution
from causica.distributions.distribution_module import DistributionModule
from causica.distributions.noise.joint import JointNoiseModule
from causica.functional_relationships.functional_relationships import FunctionalRelationships
from causica.sem.distribution_parameters_sem import DistributionParametersSEM


class SEMDistribution(td.Distribution):
    """A distribution over structural equation models.

    Samples are instances of DistributionParametersSEM.

    The distribution is essentially the same as the given adjacency distribution but with samples converted to SEMs.
    Therefore, all distribution properties such as entropy, mean and mode are given by the equivalent properties for the
    adjacency distribution.
    """

    arg_constraints: dict[str, Constraint] = {}

    def __init__(
        self,
        adjacency_dist: AdjacencyDistribution,
        noise_module: JointNoiseModule,
        functional_relationships: FunctionalRelationships,
    ):
        """
        Args:
            adjacency_dist: Distribution from which adjacency matrices are sampled to construct SEMs.
            noise_module: The noise module for any SEM of this distribution.
            functional_relationships: The functional relationship for any SEM of this distribution.
        """
        super().__init__()
        self._adjacency_dist = adjacency_dist
        self._noise_module = noise_module
        self._functional_relationships = functional_relationships

    def _create_sems(self, graphs: torch.Tensor) -> list[DistributionParametersSEM]:
        graphs = graphs.reshape(-1, *graphs.shape[-2:])
        return [
            DistributionParametersSEM(
                graph=graph,
                noise_dist=self._noise_module,
                func=self._functional_relationships,
            )
            for graph in graphs.unbind(dim=0)
        ]

    def sample(self, sample_shape: torch.Size = torch.Size()):
        graphs = self._adjacency_dist.sample(sample_shape)
        if not sample_shape:
            graphs = graphs[None, ...]
        return self._create_sems(graphs)

    def relaxed_sample(
        self, sample_shape: torch.Size = torch.Size(), temperature: float = 0.0
    ) -> list[DistributionParametersSEM]:
        graphs = self._adjacency_dist.relaxed_sample(sample_shape=sample_shape, temperature=temperature)
        return self._create_sems(graphs)

    def entropy(self) -> torch.Tensor:
        return self._adjacency_dist.entropy()

    @property
    def mean(self) -> DistributionParametersSEM:
        return DistributionParametersSEM(
            graph=self._adjacency_dist.mean,
            noise_dist=self._noise_module,
            func=self._functional_relationships,
        )

    @property
    def mode(self) -> DistributionParametersSEM:
        return DistributionParametersSEM(
            graph=self._adjacency_dist.mode,
            noise_dist=self._noise_module,
            func=self._functional_relationships,
        )

    def log_prob(self, value: DistributionParametersSEM) -> torch.Tensor:
        return self._adjacency_dist.log_prob(value.graph)


class SEMDistributionModule(DistributionModule[SEMDistribution]):
    """Represents a SEMDistribution with learnable parameters."""

    def __init__(
        self,
        adjacency_module: DistributionModule[AdjacencyDistribution],
        functional_relationships: FunctionalRelationships,
        noise_module: JointNoiseModule,
    ):
        super().__init__()
        self.adjacency_module = adjacency_module
        self.functional_relationships = functional_relationships
        self.noise_module = noise_module

    def forward(self) -> SEMDistribution:
        return SEMDistribution(
            adjacency_dist=self.adjacency_module(),
            noise_module=self.noise_module,
            functional_relationships=self.functional_relationships,
        )



import abc

import pytorch_lightning as pl
import torch
from tensordict import TensorDict

from causica.datasets.causica_dataset_format import CounterfactualWithEffects, InterventionWithEffects


class VariableSpecModule(pl.LightningModule, abc.ABC):
    """
    An Abstract Base Class for `LightningModule`s that use the VariableSpecDataModule.

    This class provides a default implementation of the `test_step` method that
    dispatches to the separate test steps for each dataloader.
    """

    def test_step(self, batch, batch_idx, dataloader_idx=0):
        """
        Dispatch to the appropriate test step based on the dataloader index.

        The dataloader indexing is handled by Lightning and is based on the order
        of the dataloaders in the `test_dataloader` method.

        This assumes that the dataloaders are in the following order:
        Test data
        Graph data
        Intervention data
        Counterfactual data

        See the superclass for *args and **kwargs conventions.
        """
        if dataloader_idx == 0:
            return self.test_step_observational(batch, batch_idx)
        elif dataloader_idx == 1:
            return self.test_step_graph(batch, batch_idx)
        elif dataloader_idx == 2:
            return self.test_step_interventions(batch, batch_idx)
        elif dataloader_idx == 3:
            return self.test_step_counterfactuals(batch, batch_idx)

    @abc.abstractmethod
    def test_step_observational(self, batch: TensorDict, *args, **kwargs):
        """Test step method for the test data."""

    @abc.abstractmethod
    def test_step_graph(self, true_adj_matrix: torch.Tensor, *args, **kwargs):
        """Test step method for the true adjacency matrix."""

    @abc.abstractmethod
    def test_step_interventions(self, interventions: InterventionWithEffects, *args, **kwargs):
        """Test step method for the interventions."""

    @abc.abstractmethod
    def test_step_counterfactuals(self, counterfactuals: CounterfactualWithEffects, *args, **kwargs):
        """Test step method for the counterfactuals."""



"""Lightning Classes for loading data in the default format used in Azure Blob Storage."""
import os
from collections import defaultdict
from functools import partial
from typing import Any, Optional

import torch
from tensordict import TensorDict
from torch.utils.data import DataLoader

from causica.datasets.causica_dataset_format import DataEnum, load_data
from causica.datasets.standardizer import JointStandardizer, fit_standardizer
from causica.datasets.tensordict_utils import identity, tensordict_shapes
from causica.datasets.variable_types import VariableTypeEnum
from causica.lightning.data_modules.deci_data_module import DECIDataModule


class VariableSpecDataModule(DECIDataModule):
    """
    Loads training and test data from fully specified paths for `variables.json` formatted data.

    This format assumes the `variables.json` specified all metadata, and that the corresponding CSV files do not have
    any header rows.

    Note:
        Uses `fsspec` to load the data in the paths. To load from a cloud storage location, make sure the relevant
        `fsspec` plugin is available and provide its corresponding scheme. Provide **storage_options for authentication.
        E.g., to load from an Azure storage account, install `adlfs` and use `az://container@storage_account/path`.

    """

    def __init__(
        self,
        root_path: str,
        batch_size: int = 128,
        dataset_name: str = "anonymous_dataset",
        normalize: bool = False,
        load_counterfactual: bool = False,
        **storage_options: dict[str, Any],
    ):
        """
        Args:
            root_path: Path to directory with causal data
            batch_size: Batch size for training and test data.
            storage_options: Storage options forwarded to `fsspec` when loading files.
            dataset_name: A name for the dataset
            load_counterfactual: Whether there is counterfactual data
            normalize: Whether to normalize the data
        """
        super().__init__()
        self.batch_size = batch_size
        self._dataset_name = dataset_name
        self.root_path = root_path
        self.batch_size = batch_size
        self.storage_options = storage_options
        self.normalize = normalize
        self.load_counterfactual = load_counterfactual

        self.normalizer: Optional[JointStandardizer] = None

    @property
    def variable_shapes(self) -> dict[str, torch.Size]:
        return _check_exists(self, "_variable_shapes")

    @property
    def variable_types(self) -> dict[str, VariableTypeEnum]:
        return _check_exists(self, "_variable_types")

    @property
    def column_names(self) -> dict[str, list[str]]:
        return _check_exists(self, "_column_names")

    @property
    def dataset_train(self) -> TensorDict:
        return _check_exists(self, "_dataset_train")

    @property
    def dataset_test(self) -> TensorDict:
        return _check_exists(self, "_dataset_test")

    @property
    def dataset_name(self) -> str:
        return self._dataset_name

    def _load_all_data(self, variables_metadata: dict):
        _load_data = partial(
            load_data, root_path=self.root_path, variables_metadata=variables_metadata, **self.storage_options
        )

        self._dataset_train = _load_data(data_enum=DataEnum.TRAIN)
        self._dataset_test = _load_data(data_enum=DataEnum.TEST)
        self.true_adj = _load_data(data_enum=DataEnum.TRUE_ADJACENCY)
        self.interventions = _load_data(data_enum=DataEnum.INTERVENTIONS)

        self.counterfactuals = None
        if self.load_counterfactual:
            self.counterfactuals = _load_data(data_enum=DataEnum.COUNTERFACTUALS)

    def prepare_data(self):
        # WARNING: Do not remove partial here. For some reason, if `load_data` is called directly from inside this data
        # module without being wrapped in partial, it's first optional `variables_metadata` argument is added to the
        # config values of this class. Upon init through the CLI, this argument then becomes interpreted as a value in
        # `storage_options`, whereby other calls to `load_data` will fail due to having two keyword arguments named
        # `variables_metadata`.
        #
        # I.e. if `load_data` is called directly from here, the lightning command line with this class as a data module
        # and `--print_config` produces:
        # ...
        # data:
        #  class_path: causica.lightning.data_modules.VariableSpecDataModule
        #  init_args:
        #    ...
        #    variables_metadata: null
        _load_data = partial(load_data, root_path=self.root_path, **self.storage_options)
        variables_metadata = _load_data(data_enum=DataEnum.VARIABLES_JSON)

        self._load_all_data(variables_metadata)

        train_keys = set(self._dataset_train.keys())
        test_keys = set(self._dataset_test.keys())
        assert (
            train_keys == test_keys
        ), f"node_names for the training and test data must match. Diff: {train_keys.symmetric_difference(test_keys)}"
        self._variable_shapes = tensordict_shapes(self._dataset_train)
        self._variable_types = {var["group_name"]: var["type"] for var in variables_metadata["variables"]}
        self._column_names = defaultdict(list)
        for variable in variables_metadata["variables"]:
            self._column_names[variable["group_name"]].append(variable["name"])

        if self.normalize:
            continuous_keys = [k for k, v in self._variable_types.items() if v == VariableTypeEnum.CONTINUOUS]
            self.normalizer = fit_standardizer(self._dataset_train.select(*continuous_keys))

            transform = self.normalizer()
            self._dataset_train = transform(self._dataset_train)
            self._dataset_test = transform(self._dataset_test)

    def train_dataloader(self):
        return DataLoader(
            dataset=self.dataset_train,
            collate_fn=identity,
            batch_size=self.batch_size,
            shuffle=True,
            drop_last=True,
        )

    def test_dataloader(self):
        test_dataloader = DataLoader(dataset=self.dataset_test, collate_fn=identity, batch_size=self.batch_size)

        dataloader_list = [
            test_dataloader,
            DataLoader(dataset=self.true_adj[None, ...]),
            DataLoader(dataset=self.interventions, collate_fn=identity, batch_size=None),
        ]

        if self.counterfactuals is not None:
            dataloader_list.append(DataLoader(dataset=self.counterfactuals, collate_fn=identity, batch_size=None))

        return dataloader_list


class CSuiteDataModule(VariableSpecDataModule):
    """CSuite data module for loading by the datasets name.

    Args:
        dataset_name: Name of the dataset to load.
        batch_size: Batch size for all datasets.
        dataset_path: Path to CSuite dataset mirror.
    """

    DEFAULT_CSUITE_PATH = "https://azuastoragepublic.blob.core.windows.net/datasets"

    def __init__(
        self,
        dataset_name: str,
        batch_size: int = 128,
        dataset_path: str = DEFAULT_CSUITE_PATH,
        load_counterfactual: bool = False,
    ):
        super().__init__(
            root_path=os.path.join(dataset_path, dataset_name),
            batch_size=batch_size,
            dataset_name=dataset_name,
            load_counterfactual=load_counterfactual,
        )


def _check_exists(obj: Any, attribute_name: str):
    """Check if an attribute exists otherwise print a message."""
    try:
        return getattr(obj, attribute_name)
    except AttributeError as exc:
        display_string = attribute_name.replace("_", " ").strip()
        raise ValueError(f"Tried to get {display_string} before data was downloaded.") from exc



import torch
import torch.nn.functional as F


def gumbel_softmax_binary(logits: torch.Tensor, tau: float, hard: bool = False) -> torch.Tensor:
    """
    Calculate the gumbel softmax for binary logits.

    Args:
        logits: A tensor of binary logits of shape (...)
        tau: Temperature of the gumbel softmax
        hard: whether samples should be hard {0, 1} or soft (in [0, 1])
    Returns:
        Tensor of shape (...)
    """
    stacked_logits = torch.stack([torch.zeros_like(logits, device=logits.device), logits], dim=-1)  # (..., 2)
    return F.gumbel_softmax(stacked_logits, tau=tau, hard=hard, dim=-1)[..., 1]  # (...) binary
